"""Core logging module for pyRevit."""
import sys
import os.path as op
import logging

#pylint: disable=W0703,C0302,C0103
from pyrevit import EXEC_PARAMS, USER_DESKTOP
from pyrevit.compat import safe_strtype
from pyrevit import PYREVIT_VERSION_APP_DIR, PYREVIT_FILE_PREFIX_STAMPED
from pyrevit import coreutils
from pyrevit.coreutils import envvars

LOG_REC_FORMAT = "%(levelname)s [%(name)s] %(message)s"
LOG_REC_FORMAT_HEADER = \
    coreutils.prepare_html_str(
        "<strong>%(levelname)s</strong> [%(name)s] %(message)s"
        )
LOG_REC_FORMAT_HEADER_NO_NAME = \
    coreutils.prepare_html_str(
        "<strong>%(levelname)s</strong>\n%(message)s"
        )
LOG_REC_FORMAT_EMOJI = "{emoji} %(levelname)s [%(name)s] %(message)s"
LOG_REC_FORMAT_FILE = "%(asctime)s %(levelname)s [%(name)s] %(message)s"
LOG_REC_FORMAT_FILE_C = "%(asctime)s %(levelname)s [<{}> %(name)s] %(message)s"

LOG_REC_FORMAT_HTML = \
    coreutils.prepare_html_str('<div class="logdefault {style}">{message}</div>')

LOG_REC_CLASS_ERROR = 'logerror'
LOG_REC_FORMAT_ERROR = \
    LOG_REC_FORMAT_HTML.format(style=LOG_REC_CLASS_ERROR,
                               message=LOG_REC_FORMAT_HEADER)

LOG_REC_CLASS_WARNING = 'logwarning'
LOG_REC_FORMAT_WARNING = \
    LOG_REC_FORMAT_HTML.format(style=LOG_REC_CLASS_WARNING,
                               message=LOG_REC_FORMAT_HEADER)

LOG_REC_CLASS_CRITICAL = 'logcritical'
LOG_REC_FORMAT_CRITICAL = \
    LOG_REC_FORMAT_HTML.format(style=LOG_REC_CLASS_CRITICAL,
                               message=LOG_REC_FORMAT_HEADER)

LOG_REC_CLASS_SUCCESS = 'logsuccess'
LOG_REC_FORMAT_SUCCESS = \
    LOG_REC_FORMAT_HTML.format(style=LOG_REC_CLASS_SUCCESS,
                               message=LOG_REC_FORMAT_HEADER_NO_NAME)

LOG_REC_CLASS_DEPRECATE = 'logdeprecate'
LOG_REC_FORMAT_DEPRECATE = \
    LOG_REC_FORMAT_HTML.format(style=LOG_REC_CLASS_DEPRECATE,
                               message=LOG_REC_FORMAT_HEADER_NO_NAME)


# Setting default global logging level
DEFAULT_LOGGING_LEVEL = logging.WARNING

# add deprecate logging level
DEPRECATE_LOG_LEVEL = 25
logging.addLevelName(DEPRECATE_LOG_LEVEL, "DEPRECATE")

# add success logging level
SUCCESS_LOG_LEVEL = 80
logging.addLevelName(SUCCESS_LOG_LEVEL, "SUCCESS")


# must be the same in this file and pyrevit/loader/runtime/envdict.cs
# this is because the csharp code hasn't been compiled when the
# logger module is imported in the other modules
envvars.set_pyrevit_env_var(envvars.LOGGING_LEVEL_ENVVAR,
                            DEFAULT_LOGGING_LEVEL)
envvars.set_pyrevit_env_var(envvars.FILELOGGING_ENVVAR,
                            False)


# Creating default file log name and status
FILE_LOG_FILENAME = '{}runtime.log'.format(PYREVIT_FILE_PREFIX_STAMPED)
FILE_LOG_FILEPATH = op.join(PYREVIT_VERSION_APP_DIR, FILE_LOG_FILENAME)
FILE_LOGGING_DEFAULT_STATE = False


# custom logger methods --------------------------------------------------------
# (for module consistency and custom adjustments)
class DispatchingFormatter(object):
    """Dispatching formatter to format by log level.

    Args:
        log_formatters (dict[int:logging.Formatter]):
            dict of level:formatter key pairs
        log_default_formatter (logging.Formatter): default formatter
    """
    def __init__(self, log_formatters, log_default_formatter):
        self._formatters = log_formatters
        self._default_formatter = log_default_formatter

    def format(self, record):
        """Format given record by log level."""
        formatter = self._formatters.get(record.levelno,
                                         self._default_formatter)
        return formatter.format(record)


class LoggerWrapper(logging.Logger):
    """Custom logging object."""
    def __init__(self, *args):
        logging.Logger.__init__(self, *args)
        self._has_errors = False
        self._filelogstate = False
        self._curlevel = DEFAULT_LOGGING_LEVEL

    def _log(self, level, msg, args, exc_info=None, extra=None): #pylint: disable=W0221
        self._has_errors = (self._has_errors or level >= logging.ERROR)

        # any report other than logging.INFO level,
        # needs to cleanup < and > character to avoid html conflict
        if not isinstance(msg, str):
            msg_str = safe_strtype(msg)
            # get rid of unicode characters
            msg_str = msg_str.encode('ascii', 'ignore')
            msg_str = msg_str.replace(op.sep, '/')
        else:
            msg_str = msg
        logging.Logger._log(self, level, msg_str, args,
                            exc_info=exc_info, extra=extra)

    def callHandlers(self, record):
        """Override logging.Logger.callHandlers."""
        for hdlr in self.handlers:
            # stream-handler only records based on current level
            if isinstance(hdlr, logging.StreamHandler) \
                    and record.levelno >= self._curlevel:
                hdlr.handle(record)
            # file-handler must record everything
            elif isinstance(hdlr, logging.FileHandler) \
                    and self._filelogstate:
                hdlr.handle(record)

    def isEnabledFor(self, level):
        """Override logging.Logger.isEnabledFor."""
        # update current logging level and file logging state
        self._filelogstate = \
            envvars.get_pyrevit_env_var(envvars.FILELOGGING_ENVVAR)
        self._curlevel = \
            envvars.get_pyrevit_env_var(envvars.LOGGING_LEVEL_ENVVAR)

        # the loader assembly sets EXEC_PARAMS.debug_mode to true if
        # user Ctrl-clicks on the button at script runtime.
        if EXEC_PARAMS.debug_mode:
            self._curlevel = logging.DEBUG

        # if file logging is disabled, return the current logging level
        # but if it's enabled, return the file logging level so the record
        # is generated and logged by file-handler. The stream-handler still
        # outputs the record based on the current logging level
        if self._filelogstate:
            return level >= logging.DEBUG

        return level >= self._curlevel

    def is_enabled_for(self, level):
        """Check if logger is enabled for level in pyRevit environment."""
        self._curlevel = \
            envvars.get_pyrevit_env_var(envvars.LOGGING_LEVEL_ENVVAR)

        # the loader assembly sets EXEC_PARAMS.debug_mode to true if
        # user Ctrl-clicks on the button at script runtime.
        if EXEC_PARAMS.debug_mode:
            self._curlevel = logging.DEBUG

        return level >= self._curlevel

    @staticmethod
    def _reset_logger_env_vars(log_level):
        envvars.set_pyrevit_env_var(envvars.LOGGING_LEVEL_ENVVAR, log_level)

    def has_errors(self):
        """Check if logger has reported any errors."""
        return self._has_errors

    def set_level(self, level):
        """Set logging level to level."""
        self._reset_logger_env_vars(level)

    def set_quiet_mode(self):
        """Activate quiet mode. All log levels are disabled."""
        self._reset_logger_env_vars(logging.CRITICAL)

    def set_verbose_mode(self):
        """Activate verbose mode. Log levels >= INFO are enabled."""
        self._reset_logger_env_vars(logging.INFO)

    def set_debug_mode(self):
        """Activate debug mode. Log levels >= DEBUG are enabled."""
        self._reset_logger_env_vars(logging.DEBUG)

    def reset_level(self):
        """Reset logging level back to default."""
        self._reset_logger_env_vars(DEFAULT_LOGGING_LEVEL)

    def get_level(self):
        """Return current logging level."""
        return envvars.get_pyrevit_env_var(envvars.LOGGING_LEVEL_ENVVAR)

    def log_parse_except(self, parsed_file, parse_ex):
        """Logs a file parsing exception.

        Args:
            parsed_file (str): File path that failed the parsing
            parse_ex (Exception): Parsing exception
        """
        err_msg = '<strong>Error while parsing file:</strong>\n{file}\n' \
                  '<strong>Error type:</strong> {type}\n' \
                  '<strong>Error Message:</strong> {errmsg}\n' \
                  '<strong>Line/Column:</strong> {lineno}/{colno}\n' \
                  '<strong>Line Text:</strong> {linetext}' \
                  .format(file=parsed_file,
                          type=parse_ex.__class__.__name__,
                          errmsg=parse_ex.msg if hasattr(parse_ex, 'msg') else "",
                          lineno=parse_ex.lineno if hasattr(parse_ex, 'lineno') else 0,
                          colno=parse_ex.offset if hasattr(parse_ex, 'offset') else 0,
                          linetext=parse_ex.text if hasattr(parse_ex, 'text') else "",
                          )
        self.error(coreutils.prepare_html_str(err_msg))

    def success(self, message, *args, **kws):
        """Log a success message.

        Args:
            message (str): success message
            *args (Any): extra agruments passed to the log function
            **kws (Any): extra agruments passed to the log function
        """
        if self.isEnabledFor(SUCCESS_LOG_LEVEL):
            # Yes, logger takes its '*args' as 'args'.
            self._log(SUCCESS_LOG_LEVEL, message, args, **kws)

    def deprecate(self, message, *args, **kws):
        """Log a deprecation message.

        Args:
            message (str): deprecation message
            *args (Any): extra agruments passed to the log function
            **kws (Any): extra agruments passed to the log function
        """
        if self.isEnabledFor(DEPRECATE_LOG_LEVEL):
            # Yes, logger takes its '*args' as 'args'.
            self._log(DEPRECATE_LOG_LEVEL, message, args, **kws)

    def dev_log(self, source, message=''):
        """Appends a message to a log file.

        Args:
            source (str): source of the message
            message (str): message to log
        """
        devlog_fname = \
            '{}.log'.format(EXEC_PARAMS.command_uniqueid or self.name)
        with open(op.join(USER_DESKTOP, devlog_fname), 'a') as devlog_file:
            devlog_file.writelines('{tstamp} [{exid}] {src}: {msg}\n'.format(
                tstamp=EXEC_PARAMS.exec_timestamp,
                exid=EXEC_PARAMS.exec_id,
                src=source,
                msg=message,
                ))


# setting up handlers and formatters -------------------------------------------
stdout_hndlr = logging.StreamHandler(sys.stdout)
# e.g [_parser] DEBUG: Can not create command.
default_formatter = logging.Formatter(LOG_REC_FORMAT)
formatters = {
    SUCCESS_LOG_LEVEL: logging.Formatter(LOG_REC_FORMAT_SUCCESS),
    logging.ERROR: logging.Formatter(LOG_REC_FORMAT_ERROR),
    logging.WARNING: logging.Formatter(LOG_REC_FORMAT_WARNING),
    logging.CRITICAL: logging.Formatter(LOG_REC_FORMAT_CRITICAL),
    DEPRECATE_LOG_LEVEL: logging.Formatter(LOG_REC_FORMAT_DEPRECATE)
    }
stdout_hndlr.setFormatter(DispatchingFormatter(formatters, default_formatter))


file_hndlr = logging.FileHandler(FILE_LOG_FILEPATH, mode='a', delay=True)
file_formatter = logging.Formatter(LOG_REC_FORMAT_FILE)
file_hndlr.setFormatter(file_formatter)


def get_stdout_hndlr():
    """Return stdout logging handler object.

    Returns:
        (logging.StreamHandler):
            configured instance of python's native stream handler
    """
    global stdout_hndlr     #pylint: disable=W0603

    return stdout_hndlr


def get_file_hndlr():
    """Return file logging handler object.

    Returns:
        (logging.FileHandler):
            configured instance of python's native stream handler
    """
    global file_hndlr       #pylint: disable=W0603

    if EXEC_PARAMS.command_mode:
        cmd_file_hndlr = logging.FileHandler(FILE_LOG_FILEPATH,
                                             mode='a', delay=True)
        logformat = LOG_REC_FORMAT_FILE_C.format(EXEC_PARAMS.command_name)
        formatter = logging.Formatter(logformat)
        cmd_file_hndlr.setFormatter(formatter)
        return cmd_file_hndlr
    else:
        return file_hndlr


# setting up public logger. this will be imported in with other modules -------
logging.setLoggerClass(LoggerWrapper)


loggers = {}


def get_logger(logger_name):
    """Register and return a logger with given name.

    Caches all registered loggers and returns the same logger object on
    second call with the same logger name.

    Args:
        logger_name (str): logger name

    Returns:
        (LoggerWrapper): logger object wrapper python's native logger

    Examples:
        ```python
        get_logger('my command')
        ```
        <LoggerWrapper ...>
    """
    if loggers.get(logger_name):
        return loggers.get(logger_name)
    else:
        logger = logging.getLogger(logger_name)    # type: LoggerWrapper
        logger.addHandler(get_stdout_hndlr())
        logger.propagate = False
        logger.addHandler(get_file_hndlr())

        loggers.update({logger_name: logger})
        return logger


def set_file_logging(status):
    """Set file logging status (enable/disable).

    Args:
        status (bool): True to enable, False to disable
    """
    envvars.set_pyrevit_env_var(envvars.FILELOGGING_ENVVAR, status)


def loggers_have_errors():
    """Check if any errors have been reported by any of registered loggers."""
    for logger in loggers.values():
        if logger.has_errors():
            return True
    return False



class ListBindingHelper(object):
 """ Provides functionality to discover a bindable list and the properties of the items contained in the list when they differ from the public properties of the object to which they bind. """
 @staticmethod
 def GetList(*__args):
  """
  GetList(dataSource: object,dataMember: str) -> object

  

   Returns an object,typically a list,from the evaluation of a specified data source and optional 

    data member.

  

  

   dataSource: The data source from which to find the list.

   dataMember: The name of the data source property that contains the list. This can be null.

   Returns: An System.Object representing the underlying list if it was found; otherwise,dataSource.

  GetList(list: object) -> object

  

   Returns a list associated with the specified data source.

  

   list: The data source to examine for its underlying list.

   Returns: An System.Object representing the underlying list if it exists; otherwise,the original data 

    source specified by list.
  """
  pass
 @staticmethod
 def GetListItemProperties(*__args):
  """
  GetListItemProperties(dataSource: object,dataMember: str,listAccessors: Array[PropertyDescriptor]) -> PropertyDescriptorCollection

  

   Returns the System.ComponentModel.PropertyDescriptorCollection that describes the properties of 

    an item type contained in the specified data member of a data source. Uses the specified 

    System.ComponentModel.PropertyDescriptor array to indicate which properties to examine.

  

  

   dataSource: The data source to be examined for property information.

   dataMember: The optional data member to be examined for property information. This can be null.

   listAccessors: The System.ComponentModel.PropertyDescriptor array describing which properties of the data 

    member to examine. This can be null.

  

   Returns: The System.ComponentModel.PropertyDescriptorCollection describing the properties of an item type 

    contained in a collection property of the specified data source.

  

  GetListItemProperties(list: object,listAccessors: Array[PropertyDescriptor]) -> PropertyDescriptorCollection

  

   Returns the System.ComponentModel.PropertyDescriptorCollection that describes the properties of 

    an item type contained in a collection property of a data source. Uses the specified 

    System.ComponentModel.PropertyDescriptor array to indicate which properties to examine.

  

  

   list: The data source to be examined for property information.

   listAccessors: The System.ComponentModel.PropertyDescriptor array describing which properties of the data 

    source to examine. This can be null.

  

   Returns: The System.ComponentModel.PropertyDescriptorCollection describing the properties of the item 

    type contained in a collection property of the data source.

  

  GetListItemProperties(list: object) -> PropertyDescriptorCollection

  

   Returns the System.ComponentModel.PropertyDescriptorCollection that describes the properties of 

    an item type contained in a specified data source,or properties of the specified data source.

  

  

   list: The data source to examine for property information.

   Returns: The System.ComponentModel.PropertyDescriptorCollection containing the properties of the items 

    contained in list,or properties of list.
  """
  pass
 @staticmethod
 def GetListItemType(*__args):
  """
  GetListItemType(dataSource: object,dataMember: str) -> Type

  

   Returns the data type of the items in the specified data source.

  

   dataSource: The data source to examine for items.

   dataMember: The optional name of the property on the data source that is to be used as the data member. This 

    can be null.

  

   Returns: For complex data binding,the System.Type of the items represented by the dataMember in the data 

    source; otherwise,the System.Type of the item in the list itself.

  

  GetListItemType(list: object) -> Type

  

   Returns the data type of the items in the specified list.

  

   list: The list to be examined for type information.

   Returns: The System.Type of the items contained in the list.
  """
  pass
 @staticmethod
 def GetListName(list,listAccessors):
  """
  GetListName(list: object,listAccessors: Array[PropertyDescriptor]) -> str

  

   Returns the name of an underlying list,given a data source and optional 

    System.ComponentModel.PropertyDescriptor array.

  

  

   list: The data source to examine for the list name.

   listAccessors: An array of System.ComponentModel.PropertyDescriptor objects to find in the data source. This 

    can be null.

  

   Returns: The name of the list in the data source,as described by listAccessors,orthe name of the data 

    source type.
  """
  pass
 __all__=[
  'GetList',
  'GetListItemProperties',
  'GetListItemType',
  'GetListName',
 ]




class AceType(Enum,IComparable,IFormattable,IConvertible):
 """
 Defines the available access control entry (ACE) types.

 

 enum AceType,values: AccessAllowed (0),AccessAllowedCallback (9),AccessAllowedCallbackObject (11),AccessAllowedCompound (4),AccessAllowedObject (5),AccessDenied (1),AccessDeniedCallback (10),AccessDeniedCallbackObject (12),AccessDeniedObject (6),MaxDefinedAceType (16),SystemAlarm (3),SystemAlarmCallback (14),SystemAlarmCallbackObject (16),SystemAlarmObject (8),SystemAudit (2),SystemAuditCallback (13),SystemAuditCallbackObject (15),SystemAuditObject (7)
 """
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==yx.__eq__(y) <==> x==yx.__eq__(y) <==> x==y """
  pass
 def __format__(self,*args):
  """ __format__(formattable: IFormattable,format: str) -> str """
  pass
 def __ge__(self,*args):
  pass
 def __gt__(self,*args):
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __le__(self,*args):
  pass
 def __lt__(self,*args):
  pass
 def __ne__(self,*args):
  pass
 def __reduce_ex__(self,*args):
  pass
 def __str__(self,*args):
  pass
 AccessAllowed=None
 AccessAllowedCallback=None
 AccessAllowedCallbackObject=None
 AccessAllowedCompound=None
 AccessAllowedObject=None
 AccessDenied=None
 AccessDeniedCallback=None
 AccessDeniedCallbackObject=None
 AccessDeniedObject=None
 MaxDefinedAceType=None
 SystemAlarm=None
 SystemAlarmCallback=None
 SystemAlarmCallbackObject=None
 SystemAlarmObject=None
 SystemAudit=None
 SystemAuditCallback=None
 SystemAuditCallbackObject=None
 SystemAuditObject=None
 value__=None




class ArrayWithOffset(object):
 """
 Encapsulates an array and an offset within the specified array.

 

 ArrayWithOffset(array: object,offset: int)
 """
 def Equals(self,obj):
  """
  Equals(self: ArrayWithOffset,obj: ArrayWithOffset) -> bool

  

   Indicates whether the specified System.Runtime.InteropServices.ArrayWithOffset object matches 

    the current instance.

  

  

   obj: An System.Runtime.InteropServices.ArrayWithOffset object to compare with this instance.

   Returns: true if the specified System.Runtime.InteropServices.ArrayWithOffset object matches the current 

    instance; otherwise,false.

  

  Equals(self: ArrayWithOffset,obj: object) -> bool

  

   Indicates whether the specified object matches the current 

    System.Runtime.InteropServices.ArrayWithOffset object.

  

  

   obj: Object to compare with this instance.

   Returns: true if the object matches this System.Runtime.InteropServices.ArrayWithOffset; otherwise,false.
  """
  pass
 def GetArray(self):
  """
  GetArray(self: ArrayWithOffset) -> object

  

   Returns the managed array referenced by this System.Runtime.InteropServices.ArrayWithOffset.

   Returns: The managed array this instance references.
  """
  pass
 def GetHashCode(self):
  """
  GetHashCode(self: ArrayWithOffset) -> int

  

   Returns a hash code for this value type.

   Returns: The hash code for this instance.
  """
  pass
 def GetOffset(self):
  """
  GetOffset(self: ArrayWithOffset) -> int

  

   Returns the offset provided when this System.Runtime.InteropServices.ArrayWithOffset was 

    constructed.

  

   Returns: The offset for this instance.
  """
  pass
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==y """
  pass
 @staticmethod
 def __new__(self,array,offset):
  """
  __new__(cls: type,array: object,offset: int)

  __new__[ArrayWithOffset]() -> ArrayWithOffset
  """
  pass
 def __ne__(self,*args):
  pass



class DataGridViewAutoSizeColumnsModeEventArgs(EventArgs):
 """
 Provides data for the System.Windows.Forms.DataGridView.AutoSizeColumnsModeChanged event.

 

 DataGridViewAutoSizeColumnsModeEventArgs(previousModes: Array[DataGridViewAutoSizeColumnMode])
 """
 @staticmethod
 def __new__(self,previousModes):
  """ __new__(cls: type,previousModes: Array[DataGridViewAutoSizeColumnMode]) """
  pass
 PreviousModes=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets an array of the previous values of the column System.Windows.Forms.DataGridViewColumn.AutoSizeMode properties.



Get: PreviousModes(self: DataGridViewAutoSizeColumnsModeEventArgs) -> Array[DataGridViewAutoSizeColumnMode]



"""





class DownloadProgressChangedEventArgs(ProgressChangedEventArgs):
 """ Provides data for the System.Net.WebClient.DownloadProgressChanged event of a System.Net.WebClient. """
 BytesReceived=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the number of bytes received.



Get: BytesReceived(self: DownloadProgressChangedEventArgs) -> Int64



"""

 TotalBytesToReceive=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the total number of bytes in a System.Net.WebClient data download operation.



Get: TotalBytesToReceive(self: DownloadProgressChangedEventArgs) -> Int64



"""





class IEquatable:
 # no doc
 def Equals(self,other):
  """
  Equals(self: IEquatable[T],other: T) -> bool

  

   Indicates whether the current object is equal to another object of the same type.

  

   other: An object to compare with this object.

   Returns: true if the current object is equal to the other parameter; otherwise,false.
  """
  pass
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==y """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass



"""IxMIlia.Dxf assembly import."""
# pylint: skip-file
import os.path as op
from pyrevit import clr, BIN_DIR

clr.AddReferenceToFileAndPath(
    op.join(BIN_DIR, 'IxMilia.Dxf')
    )

from IxMilia import Dxf



class ItemCheckedEventArgs(EventArgs):
 """
 Provides data for the System.Windows.Forms.ListView.ItemChecked event of the System.Windows.Forms.ListView control.

 

 ItemCheckedEventArgs(item: ListViewItem)
 """
 def __getitem__(self,*args):
  """ x.__getitem__(y) <==> x[y] """
  pass
 @staticmethod
 def __new__(self,item):
  """ __new__(cls: type,item: ListViewItem) """
  pass
 Item=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the System.Windows.Forms.ListViewItem whose checked state is changing.



Get: Item(self: ItemCheckedEventArgs) -> ListViewItem



"""





class WebExceptionStatus(Enum,IComparable,IFormattable,IConvertible):
 """
 Defines status codes for the System.Net.WebException class.

 

 enum WebExceptionStatus,values: CacheEntryNotFound (18),ConnectFailure (2),ConnectionClosed (8),KeepAliveFailure (12),MessageLengthLimitExceeded (17),NameResolutionFailure (1),Pending (13),PipelineFailure (5),ProtocolError (7),ProxyNameResolutionFailure (15),ReceiveFailure (3),RequestCanceled (6),RequestProhibitedByCachePolicy (19),RequestProhibitedByProxy (20),SecureChannelFailure (10),SendFailure (4),ServerProtocolViolation (11),Success (0),Timeout (14),TrustFailure (9),UnknownError (16)
 """
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==yx.__eq__(y) <==> x==yx.__eq__(y) <==> x==y """
  pass
 def __format__(self,*args):
  """ __format__(formattable: IFormattable,format: str) -> str """
  pass
 def __ge__(self,*args):
  pass
 def __gt__(self,*args):
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __le__(self,*args):
  pass
 def __lt__(self,*args):
  pass
 def __ne__(self,*args):
  pass
 def __reduce_ex__(self,*args):
  pass
 def __str__(self,*args):
  pass
 CacheEntryNotFound=None
 ConnectFailure=None
 ConnectionClosed=None
 KeepAliveFailure=None
 MessageLengthLimitExceeded=None
 NameResolutionFailure=None
 Pending=None
 PipelineFailure=None
 ProtocolError=None
 ProxyNameResolutionFailure=None
 ReceiveFailure=None
 RequestCanceled=None
 RequestProhibitedByCachePolicy=None
 RequestProhibitedByProxy=None
 SecureChannelFailure=None
 SendFailure=None
 ServerProtocolViolation=None
 Success=None
 Timeout=None
 TrustFailure=None
 UnknownError=None
 value__=None




class IFormattable:
 """ Provides functionality to format the value of an object into a string representation. """
 def ToString(self,format,formatProvider):
  """
  ToString(self: IFormattable,format: str,formatProvider: IFormatProvider) -> str

  

   Formats the value of the current instance using the specified format.

  

   format: The format to use.-or- A null reference (Nothing in Visual Basic) to use the default format 

    defined for the type of the System.IFormattable implementation.

  

   formatProvider: The provider to use to format the value.-or- A null reference (Nothing in Visual Basic) to 

    obtain the numeric format information from the current locale setting of the operating system.

  

   Returns: The value of the current instance in the specified format.
  """
  pass
 def __format__(self,*args):
  """ __format__(formattable: IFormattable,format: str) -> str """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass



class HtmlElementEventArgs(EventArgs):
 """ Provides data for the events defined on System.Windows.Forms.HtmlDocument and System.Windows.Forms.HtmlElement. """
 AltKeyPressed=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Indicates whether the ALT key was pressed when this event occurred.



Get: AltKeyPressed(self: HtmlElementEventArgs) -> bool



"""

 BubbleEvent=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets a value indicating whether the current event bubbles up through the element hierarchy of the HTML Document Object Model.



Get: BubbleEvent(self: HtmlElementEventArgs) -> bool



Set: BubbleEvent(self: HtmlElementEventArgs)=value

"""

 ClientMousePosition=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the position of the mouse cursor in the document's client area.



Get: ClientMousePosition(self: HtmlElementEventArgs) -> Point



"""

 CtrlKeyPressed=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Indicates whether the CTRL key was pressed when this event occurred.



Get: CtrlKeyPressed(self: HtmlElementEventArgs) -> bool



"""

 EventType=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the name of the event that was raised.



Get: EventType(self: HtmlElementEventArgs) -> str



"""

 FromElement=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the System.Windows.Forms.HtmlElement the mouse pointer is moving away from.



Get: FromElement(self: HtmlElementEventArgs) -> HtmlElement



"""

 KeyPressedCode=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the ASCII value of the keyboard character typed in a System.Windows.Forms.HtmlElement.KeyPress,System.Windows.Forms.HtmlElement.KeyDown,or System.Windows.Forms.HtmlElement.KeyUp event.



Get: KeyPressedCode(self: HtmlElementEventArgs) -> int



"""

 MouseButtonsPressed=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the mouse button that was clicked during a System.Windows.Forms.HtmlElement.MouseDown or System.Windows.Forms.HtmlElement.MouseUp event.



Get: MouseButtonsPressed(self: HtmlElementEventArgs) -> MouseButtons



"""

 MousePosition=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the position of the mouse cursor relative to a relatively positioned parent element.



Get: MousePosition(self: HtmlElementEventArgs) -> Point



"""

 OffsetMousePosition=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the position of the mouse cursor relative to the element that raises the event.



Get: OffsetMousePosition(self: HtmlElementEventArgs) -> Point



"""

 ReturnValue=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the return value of the handled event.



Get: ReturnValue(self: HtmlElementEventArgs) -> bool



Set: ReturnValue(self: HtmlElementEventArgs)=value

"""

 ShiftKeyPressed=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Indicates whether the SHIFT key was pressed when this event occurred.



Get: ShiftKeyPressed(self: HtmlElementEventArgs) -> bool



"""

 ToElement=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the System.Windows.Forms.HtmlElement toward which the user is moving the mouse pointer.



Get: ToElement(self: HtmlElementEventArgs) -> HtmlElement



"""





"""Handle reading and parsing, writing and saving of all user configurations.

This module handles the reading and writing of the pyRevit configuration files.
It's been used extensively by pyRevit sub-modules. user_config is
set up automatically in the global scope by this module and can be imported
into scripts and other modules to access the default configurations.

All other modules use this module to query user config.

Examples:
    ```python
    from pyrevit.userconfig import user_config
    user_config.add_section('newsection')
    user_config.newsection.property = value
    user_config.newsection.get_option('property', default_value)
    user_config.save_changes()
    ```


The user_config object is also the destination for reading and writing
configuration by pyRevit scripts through :func:`get_config` of
:mod:`pyrevit.script` module. Here is the function source:

.. literalinclude:: ../../pyrevitlib/pyrevit/script.py
    :pyobject: get_config

Examples:
    ```python
    from pyrevit import script
    cfg = script.get_config()
    cfg.property = value
    cfg.get_option('property', default_value)
    script.save_config()
    ```
"""
#pylint: disable=C0103,C0413,W0703
import os
import os.path as op

from pyrevit import EXEC_PARAMS, HOME_DIR, HOST_APP
from pyrevit import PyRevitException
from pyrevit import EXTENSIONS_DEFAULT_DIR, THIRDPARTY_EXTENSIONS_DEFAULT_DIR
from pyrevit import PYREVIT_ALLUSER_APP_DIR, PYREVIT_APP_DIR
from pyrevit.compat import winreg as wr

from pyrevit.labs import PyRevit

from pyrevit import coreutils
from pyrevit.coreutils import appdata
from pyrevit.coreutils import configparser
from pyrevit.coreutils import logger
from pyrevit.versionmgr import upgrade
# pylint: disable=C0103,C0413,W0703
DEFAULT_CSV_SEPARATOR = ','


mlogger = logger.get_logger(__name__)


CONSTS = PyRevit.PyRevitConsts


class PyRevitConfig(configparser.PyRevitConfigParser):
    """Provide read/write access to pyRevit configuration.

    Args:
        cfg_file_path (str): full path to config file to be used.
        config_type (str): type of config file

    Examples:
        ```python
        cfg = PyRevitConfig(cfg_file_path)
        cfg.add_section('sectionname')
        cfg.sectionname.property = value
        cfg.sectionname.get_option('property', default_value)
        cfg.save_changes()
        ```
    """

    def __init__(self, cfg_file_path=None, config_type='Unknown'):
        """Load settings from provided config file and setup parser."""
        # try opening and reading config file in order.
        super(PyRevitConfig, self).__init__(cfg_file_path=cfg_file_path)

        # set log mode on the logger module based on
        # user settings (overriding the defaults)
        self._update_env()
        self._admin = config_type == 'Admin'
        self.config_type = config_type

    def _update_env(self):
        # update the debug level based on user config
        mlogger.reset_level()

        try:
            # first check to see if command is not in forced debug mode
            if not EXEC_PARAMS.debug_mode:
                if self.core.debug:
                    mlogger.set_debug_mode()
                    mlogger.debug('Debug mode is enabled in user settings.')
                elif self.core.verbose:
                    mlogger.set_verbose_mode()

            logger.set_file_logging(self.core.filelogging)
        except Exception as env_update_err:
            mlogger.debug('Error updating env variable per user config. | %s',
                          env_update_err)

    @property
    def config_file(self):
        """Current config file path."""
        return self._cfg_file_path

    @property
    def environment(self):
        """Environment section."""
        if not self.has_section(CONSTS.EnvConfigsSectionName):
            self.add_section(CONSTS.EnvConfigsSectionName)
        return self.get_section(CONSTS.EnvConfigsSectionName)

    @property
    def core(self):
        """Core section."""
        if not self.has_section(CONSTS.ConfigsCoreSection):
            self.add_section(CONSTS.ConfigsCoreSection)
        return self.get_section(CONSTS.ConfigsCoreSection)

    @property
    def routes(self):
        """Routes section."""
        if not self.has_section(CONSTS.ConfigsRoutesSection):
            self.add_section(CONSTS.ConfigsRoutesSection)
        return self.get_section(CONSTS.ConfigsRoutesSection)

    @property
    def telemetry(self):
        """Telemetry section."""
        if not self.has_section(CONSTS.ConfigsTelemetrySection):
            self.add_section(CONSTS.ConfigsTelemetrySection)
        return self.get_section(CONSTS.ConfigsTelemetrySection)

    @property
    def bin_cache(self):
        """"Whether to use the cache for extensions."""
        return self.core.get_option(
            CONSTS.ConfigsBinaryCacheKey,
            default_value=CONSTS.ConfigsBinaryCacheDefault,
        )

    @bin_cache.setter
    def bin_cache(self, state):
        self.core.set_option(
            CONSTS.ConfigsBinaryCacheKey,
            value=state
        )

    @property
    def check_updates(self):
        """Whether to check for updates."""
        return self.core.get_option(
            CONSTS.ConfigsCheckUpdatesKey,
            default_value=CONSTS.ConfigsCheckUpdatesDefault,
        )

    @check_updates.setter
    def check_updates(self, state):
        self.core.set_option(
            CONSTS.ConfigsCheckUpdatesKey,
            value=state
        )

    @property
    def auto_update(self):
        """Whether to automatically update pyRevit."""
        return self.core.get_option(
            CONSTS.ConfigsAutoUpdateKey,
            default_value=CONSTS.ConfigsAutoUpdateDefault,
        )

    @auto_update.setter
    def auto_update(self, state):
        self.core.set_option(
            CONSTS.ConfigsAutoUpdateKey,
            value=state
        )

    @property
    def rocket_mode(self):
        """Whether to enable rocket mode."""
        return self.core.get_option(
            CONSTS.ConfigsRocketModeKey,
            default_value=CONSTS.ConfigsRocketModeDefault,
        )

    @rocket_mode.setter
    def rocket_mode(self, state):
        self.core.set_option(
            CONSTS.ConfigsRocketModeKey,
            value=state
        )

    @property
    def log_level(self):
        """Logging level."""
        if self.core.get_option(
                CONSTS.ConfigsDebugKey,
                default_value=CONSTS.ConfigsDebugDefault,
            ):
            return PyRevit.PyRevitLogLevels.Debug
        elif self.core.get_option(
                CONSTS.ConfigsVerboseKey,
                default_value=CONSTS.ConfigsVerboseDefault,
            ):
            return PyRevit.PyRevitLogLevels.Verbose
        return PyRevit.PyRevitLogLevels.Quiet

    @log_level.setter
    def log_level(self, state):
        if state == PyRevit.PyRevitLogLevels.Debug:
            self.core.set_option(CONSTS.ConfigsDebugKey, True)
            self.core.set_option(CONSTS.ConfigsVerboseKey, True)
        elif state == PyRevit.PyRevitLogLevels.Verbose:
            self.core.set_option(CONSTS.ConfigsDebugKey, False)
            self.core.set_option(CONSTS.ConfigsVerboseKey, True)
        else:
            self.core.set_option(CONSTS.ConfigsDebugKey, False)
            self.core.set_option(CONSTS.ConfigsVerboseKey, False)

    @property
    def file_logging(self):
        """Whether to enable file logging."""
        return self.core.get_option(
            CONSTS.ConfigsFileLoggingKey,
            default_value=CONSTS.ConfigsFileLoggingDefault,
        )

    @file_logging.setter
    def file_logging(self, state):
        self.core.set_option(
            CONSTS.ConfigsFileLoggingKey,
            value=state
        )

    @property
    def startuplog_timeout(self):
        """Timeout for the startup log."""
        return self.core.get_option(
            CONSTS.ConfigsStartupLogTimeoutKey,
            default_value=CONSTS.ConfigsStartupLogTimeoutDefault,
        )

    @startuplog_timeout.setter
    def startuplog_timeout(self, timeout):
        self.core.set_option(
            CONSTS.ConfigsStartupLogTimeoutKey,
            value=timeout
        )

    @property
    def required_host_build(self):
        """Host build required to run the commands."""
        return self.core.get_option(
            CONSTS.ConfigsRequiredHostBuildKey,
            default_value="",
        )

    @required_host_build.setter
    def required_host_build(self, buildnumber):
        self.core.set_option(
            CONSTS.ConfigsRequiredHostBuildKey,
            value=buildnumber
        )

    @property
    def min_host_drivefreespace(self):
        """Minimum free space for running the commands."""
        return self.core.get_option(
            CONSTS.ConfigsMinDriveSpaceKey,
            default_value=CONSTS.ConfigsMinDriveSpaceDefault,
        )

    @min_host_drivefreespace.setter
    def min_host_drivefreespace(self, freespace):
        self.core.set_option(
            CONSTS.ConfigsMinDriveSpaceKey,
            value=freespace
        )

    @property
    def load_beta(self):
        """Whether to load commands in beta."""
        return self.core.get_option(
            CONSTS.ConfigsLoadBetaKey,
            default_value=CONSTS.ConfigsLoadBetaDefault,
        )

    @load_beta.setter
    def load_beta(self, state):
        self.core.set_option(
            CONSTS.ConfigsLoadBetaKey,
            value=state
        )

    @property
    def output_close_others(self):
        """Whether to close other output windows."""
        return self.core.get_option(
            CONSTS.ConfigsCloseOtherOutputsKey,
            default_value=CONSTS.ConfigsCloseOtherOutputsDefault,
        )

    @output_close_others.setter
    def output_close_others(self, state):
        self.core.set_option(
            CONSTS.ConfigsCloseOtherOutputsKey,
            value=state
        )

    @property
    def output_close_mode_enum(self):
        """Output window closing mode as enum (CurrentCommand | CloseAll)."""
        value = self.core.get_option(
            CONSTS.ConfigsCloseOutputModeKey,
            default_value=CONSTS.ConfigsCloseOutputModeDefault,
        )
        if not value:
            value = CONSTS.ConfigsCloseOutputModeDefault

        value_lc = str(value).lower()

        if value_lc == str(CONSTS.ConfigsCloseOutputModeCloseAll).lower():
            return PyRevit.OutputCloseMode.CloseAll
        else:
            return PyRevit.OutputCloseMode.CurrentCommand

    @output_close_mode_enum.setter
    def output_close_mode_enum(self, mode):
        """Store string in INI, mapped from enum."""
        if mode == PyRevit.OutputCloseMode.CloseAll:
            self.core.set_option(
                CONSTS.ConfigsCloseOutputModeKey,
                value=CONSTS.ConfigsCloseOutputModeCloseAll
            )
        else:
            self.core.set_option(
                CONSTS.ConfigsCloseOutputModeKey,
                value=CONSTS.ConfigsCloseOutputModeCurrentCommand
            )

    @property
    def cpython_engine_version(self):
        """CPython engine version to use."""
        return self.core.get_option(
            CONSTS.ConfigsCPythonEngineKey,
            default_value=CONSTS.ConfigsCPythonEngineDefault,
        )

    @cpython_engine_version.setter
    def cpython_engine_version(self, version):
        self.core.set_option(
            CONSTS.ConfigsCPythonEngineKey,
            value=int(version)
        )

    @property
    def user_locale(self):
        """User locale."""
        return self.core.get_option(
            CONSTS.ConfigsLocaleKey,
            default_value="",
        )

    @user_locale.setter
    def user_locale(self, local_code):
        self.core.set_option(
            CONSTS.ConfigsLocaleKey,
            value=local_code
        )

    @property
    def output_stylesheet(self):
        """Stylesheet used for output."""
        return self.core.get_option(
            CONSTS.ConfigsOutputStyleSheet,
            default_value="",
        )

    @output_stylesheet.setter
    def output_stylesheet(self, stylesheet_filepath):
        if stylesheet_filepath:
            self.core.set_option(
                CONSTS.ConfigsOutputStyleSheet,
                value=stylesheet_filepath
            )
        else:
            self.core.remove_option(CONSTS.ConfigsOutputStyleSheet)

    @property
    def routes_host(self):
        """Routes API host."""
        return self.routes.get_option(
            CONSTS.ConfigsRoutesHostKey,
            default_value=CONSTS.ConfigsRoutesHostDefault,
        )

    @routes_host.setter
    def routes_host(self, routes_host):
        self.routes.set_option(
            CONSTS.ConfigsRoutesHostKey,
            value=routes_host
        )

    @property
    def routes_port(self):
        """API routes port."""
        return self.routes.get_option(
            CONSTS.ConfigsRoutesPortKey,
            default_value=CONSTS.ConfigsRoutesPortDefault,
        )

    @routes_port.setter
    def routes_port(self, port):
        self.routes.set_option(
            CONSTS.ConfigsRoutesPortKey,
            value=port
        )

    @property
    def load_core_api(self):
        """Whether to load pyRevit core api."""
        return self.routes.get_option(
            CONSTS.ConfigsLoadCoreAPIKey,
            default_value=CONSTS.ConfigsConfigsLoadCoreAPIDefault,
        )

    @load_core_api.setter
    def load_core_api(self, state):
        self.routes.set_option(
            CONSTS.ConfigsLoadCoreAPIKey,
            value=state
        )

    @property
    def telemetry_utc_timestamp(self):
        """Whether to use UTC timestamps in telemetry."""
        return self.telemetry.get_option(
            CONSTS.ConfigsTelemetryUTCTimestampsKey,
            default_value=CONSTS.ConfigsTelemetryUTCTimestampsDefault,
        )

    @telemetry_utc_timestamp.setter
    def telemetry_utc_timestamp(self, state):
        self.telemetry.set_option(
            CONSTS.ConfigsTelemetryUTCTimestampsKey,
            value=state
        )

    @property
    def telemetry_status(self):
        """Telemetry status."""
        return self.telemetry.get_option(
            CONSTS.ConfigsTelemetryStatusKey,
            default_value=CONSTS.ConfigsTelemetryStatusDefault,
        )

    @telemetry_status.setter
    def telemetry_status(self, state):
        self.telemetry.set_option(
            CONSTS.ConfigsTelemetryStatusKey,
            value=state
        )

    @property
    def telemetry_file_dir(self):
        """Telemetry file directory."""
        return self.telemetry.get_option(
            CONSTS.ConfigsTelemetryFileDirKey,
            default_value="",
        )

    @telemetry_file_dir.setter
    def telemetry_file_dir(self, filepath):
        self.telemetry.set_option(
            CONSTS.ConfigsTelemetryFileDirKey,
            value=filepath
        )

    @property
    def telemetry_server_url(self):
        """Telemetry server URL."""
        return self.telemetry.get_option(
            CONSTS.ConfigsTelemetryServerUrlKey,
            default_value="",
        )

    @telemetry_server_url.setter
    def telemetry_server_url(self, server_url):
        self.telemetry.set_option(
            CONSTS.ConfigsTelemetryServerUrlKey,
            value=server_url
        )

    @property
    def telemetry_include_hooks(self):
        """Whether to include hooks in telemetry."""
        return self.telemetry.get_option(
            CONSTS.ConfigsTelemetryIncludeHooksKey,
            default_value=CONSTS.ConfigsTelemetryIncludeHooksDefault,
        )

    @telemetry_include_hooks.setter
    def telemetry_include_hooks(self, state):
        self.telemetry.set_option(
            CONSTS.ConfigsTelemetryIncludeHooksKey,
            value=state
        )

    @property
    def apptelemetry_status(self):
        """Telemetry status."""
        return self.telemetry.get_option(
            CONSTS.ConfigsAppTelemetryStatusKey,
            default_value=CONSTS.ConfigsAppTelemetryStatusDefault,
        )

    @apptelemetry_status.setter
    def apptelemetry_status(self, state):
        self.telemetry.set_option(
            CONSTS.ConfigsAppTelemetryStatusKey,
            value=state
        )

    @property
    def apptelemetry_server_url(self):
        """App telemetry server URL."""
        return self.telemetry.get_option(
            CONSTS.ConfigsAppTelemetryServerUrlKey,
            default_value="",
        )

    @apptelemetry_server_url.setter
    def apptelemetry_server_url(self, server_url):
        self.telemetry.set_option(
            CONSTS.ConfigsAppTelemetryServerUrlKey,
            value=server_url
        )

    @property
    def apptelemetry_event_flags(self):
        """Telemetry event flags."""
        return self.telemetry.get_option(
            CONSTS.ConfigsAppTelemetryEventFlagsKey,
            default_value="",
        )

    @apptelemetry_event_flags.setter
    def apptelemetry_event_flags(self, flags):
        self.telemetry.set_option(
            CONSTS.ConfigsAppTelemetryEventFlagsKey,
            value=flags
        )

    @property
    def user_can_update(self):
        """Whether the user can update pyRevit repos."""
        return self.core.get_option(
            CONSTS.ConfigsUserCanUpdateKey,
            default_value=CONSTS.ConfigsUserCanUpdateDefault,
        )

    @user_can_update.setter
    def user_can_update(self, state):
        self.core.set_option(
            CONSTS.ConfigsUserCanUpdateKey,
            value=state
        )

    @property
    def user_can_extend(self):
        """Whether the user can manage pyRevit Extensions."""
        return self.core.get_option(
            CONSTS.ConfigsUserCanExtendKey,
            default_value=CONSTS.ConfigsUserCanExtendDefault,
        )

    @user_can_extend.setter
    def user_can_extend(self, state):
        self.core.set_option(
            CONSTS.ConfigsUserCanExtendKey,
            value=state
        )

    @property
    def user_can_config(self):
        """Whether the user can access the configuration."""
        return self.core.get_option(
            CONSTS.ConfigsUserCanConfigKey,
            default_value=CONSTS.ConfigsUserCanConfigDefault,
        )

    @user_can_config.setter
    def user_can_config(self, state):
        self.core.set_option(
            CONSTS.ConfigsUserCanConfigKey,
            value=state
        )

    @property
    def colorize_docs(self):
        """Whether to enable the document colorizer."""
        return self.core.get_option(
            CONSTS.ConfigsColorizeDocsKey,
            default_value=CONSTS.ConfigsColorizeDocsDefault,
        )

    @colorize_docs.setter
    def colorize_docs(self, state):
        self.core.set_option(
            CONSTS.ConfigsColorizeDocsKey,
            value=state
        )

    @property
    def tooltip_debug_info(self):
        """Whether to append debug info on tooltips."""
        return self.core.get_option(
            CONSTS.ConfigsAppendTooltipExKey,
            default_value=CONSTS.ConfigsAppendTooltipExDefault,
        )

    @tooltip_debug_info.setter
    def tooltip_debug_info(self, state):
        self.core.set_option(
            CONSTS.ConfigsAppendTooltipExKey,
            value=state
        )

    @property
    def routes_server(self):
        """Whether the server routes are enabled."""
        return self.routes.get_option(
            CONSTS.ConfigsRoutesServerKey,
            default_value=CONSTS.ConfigsRoutesServerDefault,
        )

    @routes_server.setter
    def routes_server(self, state):
        self.routes.set_option(
            CONSTS.ConfigsRoutesServerKey,
            value=state
        )

    @property
    def respect_language_direction(self):
        """Whether the system respects the language direction."""
        return False

    @respect_language_direction.setter
    def respect_language_direction(self, state):
        pass

    def get_config_version(self):
        """Return version of config file used for change detection.

        Returns:
            (str): hash of the config file
        """
        return self.get_config_file_hash()

    def get_thirdparty_ext_root_dirs(self, include_default=True):
        """Return a list of external extension directories set by the user.

        Returns:
            (list[str]): External user extension directories.
        """
        dir_list = set()
        if include_default:
            # add default ext path
            dir_list.add(THIRDPARTY_EXTENSIONS_DEFAULT_DIR)
        try:
            dir_list.update([
                op.expandvars(op.normpath(x))
                for x in self.core.get_option(
                    CONSTS.ConfigsUserExtensionsKey,
                    default_value=[]
                )])
        except Exception as read_err:
            mlogger.error('Error reading list of user extension folders. | %s',
                          read_err)

        return [x for x in dir_list if op.exists(x)]

    def get_ext_root_dirs(self):
        """Return a list of all extension directories.

        Returns:
            (list[str]): user extension directories.

        """
        dir_list = set()
        if op.exists(EXTENSIONS_DEFAULT_DIR):
            dir_list.add(EXTENSIONS_DEFAULT_DIR)
        dir_list.update(self.get_thirdparty_ext_root_dirs())
        return list(dir_list)

    def get_ext_sources(self):
        """Return a list of extension definition source files."""
        ext_sources = self.environment.get_option(
            CONSTS.EnvConfigsExtensionLookupSourcesKey,
            default_value=[],
        )
        return list(set(ext_sources))

    def set_thirdparty_ext_root_dirs(self, path_list):
        """Updates list of external extension directories in config file.

        Args:
            path_list (list[str]): list of external extension paths
        """
        for ext_path in path_list:
            if not op.exists(ext_path):
                raise PyRevitException("Path \"%s\" does not exist." % ext_path)

        try:
            self.core.userextensions = \
                [op.normpath(x) for x in path_list]
        except Exception as write_err:
            mlogger.error('Error setting list of user extension folders. | %s',
                          write_err)

    def get_current_attachment(self):
        """Return current pyRevit attachment."""
        try:
            return PyRevit.PyRevitAttachments.GetAttached(int(HOST_APP.version))
        except PyRevitException as ex:
            mlogger.error('Error getting current attachment. | %s', ex)

    def get_active_cpython_engine(self):
        """Return active cpython engine."""
        # try to find attachment and get engines from the clone
        attachment = self.get_current_attachment()
        if attachment and attachment.Clone:
            clone = attachment.Clone
        else:
            # if can not find attachment, instantiate a temp clone
            try:
                clone = PyRevit.PyRevitClone(clonePath=HOME_DIR)
            except Exception as cEx:
                mlogger.debug('Can not create clone from path: %s', str(cEx))
                clone = None
        # find cpython engines
        engines = clone.GetCPythonEngines() if clone else []
        cpy_engines_dict = {x.Version: x for x in engines}
        mlogger.debug('cpython engines dict: %s', cpy_engines_dict)

        if not cpy_engines_dict:
            mlogger.error(
                'Can not determine cpython engines for current attachment: %s',
                attachment
            )
            return None
        # grab cpython engine configured to be used by user
        try:
            cpyengine_ver = int(self.cpython_engine_version)
        except (ValueError, TypeError):
            cpyengine_ver = 000

        try:
            return cpy_engines_dict[cpyengine_ver]
        except KeyError:
            # return the latest cpython engine
            return max(cpy_engines_dict.values(), key=lambda x: x.Version.Version)

    def set_active_cpython_engine(self, pyrevit_engine):
        """Set the active CPython engine.

        Args:
            pyrevit_engine (PyRevitEngine): python engine to set as active
        """
        self.cpython_engine_version = pyrevit_engine.Version

    @property
    def is_readonly(self):
        """bool: whether the config is read only."""
        return self._admin

    def save_changes(self):
        """Save user config into associated config file."""
        if not self._admin and self.config_file:
            try:
                super(PyRevitConfig, self).save()
            except Exception as save_err:
                mlogger.error('Can not save user config to: %s | %s',
                              self.config_file, save_err)

            # adjust environment per user configurations
            self._update_env()
        else:
            mlogger.debug('Config is in admin mode. Skipping save.')

    @staticmethod
    def get_list_separator():
        """Get list separator defined in user os regional settings."""
        intkey = coreutils.get_reg_key(wr.HKEY_CURRENT_USER,
                                       r'Control Panel\International')
        if intkey:
            try:
                return wr.QueryValueEx(intkey, 'sList')[0]
            except Exception:
                return DEFAULT_CSV_SEPARATOR


def find_config_file(target_path):
    """Find config file in target path."""
    return PyRevit.PyRevitConsts.FindConfigFileInDirectory(target_path)


def verify_configs(config_file_path=None):
    """Create a user settings file.

    if config_file_path is not provided, configs will be in memory only

    Args:
        config_file_path (str, optional): config file full name and path

    Returns:
        (pyrevit.userconfig.PyRevitConfig): pyRevit config file handler
    """
    if config_file_path:
        mlogger.debug('Creating default config file at: %s', config_file_path)
        coreutils.touch(config_file_path)

    try:
        parser = PyRevitConfig(cfg_file_path=config_file_path)
    except Exception as read_err:
        # can not create default user config file under appdata folder
        mlogger.warning('Can not create config file under: %s | %s',
                        config_file_path, read_err)
        parser = PyRevitConfig()

    return parser


LOCAL_CONFIG_FILE = ADMIN_CONFIG_FILE = USER_CONFIG_FILE = CONFIG_FILE = ''
user_config = None

# location for default pyRevit config files
LOCAL_CONFIG_FILE = find_config_file(HOME_DIR)
ADMIN_CONFIG_FILE = find_config_file(PYREVIT_ALLUSER_APP_DIR)
USER_CONFIG_FILE = find_config_file(PYREVIT_APP_DIR)

# decide which config file to use
# check if a config file is inside the repo. for developers config override
if LOCAL_CONFIG_FILE:
    CONFIG_TYPE = 'Local'
    CONFIG_FILE = LOCAL_CONFIG_FILE

# check to see if there is any config file provided by admin
elif ADMIN_CONFIG_FILE \
        and os.access(ADMIN_CONFIG_FILE, os.W_OK) \
        and not USER_CONFIG_FILE:
    # if yes, copy that and use as default
    # if admin config file is writable it means it is provided
    # to bootstrap the first pyRevit run
    try:
        CONFIG_TYPE = 'Seed'
        # make a local copy if one does not exist
        PyRevit.PyRevitConfigs.SetupConfig(ADMIN_CONFIG_FILE)
        CONFIG_FILE = find_config_file(PYREVIT_APP_DIR)
    except Exception as adminEx:
        # if init operation failed, make a new config file
        CONFIG_TYPE = 'New'
        # setup config file name and path
        CONFIG_FILE = appdata.get_universal_data_file(file_id='config',
                                                        file_ext='ini')
        mlogger.warning(
            'Failed to initialize config from seed file at %s\n'
            'Using default config file',
            ADMIN_CONFIG_FILE
        )

# unless it's locked. then read that config file and set admin-mode
elif ADMIN_CONFIG_FILE \
        and not os.access(ADMIN_CONFIG_FILE, os.W_OK):
    CONFIG_TYPE = 'Admin'
    CONFIG_FILE = ADMIN_CONFIG_FILE

# if a config file is available for user use that
elif USER_CONFIG_FILE:
    CONFIG_TYPE = 'User'
    CONFIG_FILE = USER_CONFIG_FILE

# if nothing can be found, make a new one
else:
    CONFIG_TYPE = 'New'
    # setup config file name and path
    CONFIG_FILE = appdata.get_universal_data_file(file_id='config',
                                                    file_ext='ini')

mlogger.debug('Using %s config file: %s', CONFIG_TYPE, CONFIG_FILE)

# read config, or setup default config file if not available
# this pushes reading settings at first import of this module.
try:
    verify_configs(CONFIG_FILE)
    user_config = PyRevitConfig(cfg_file_path=CONFIG_FILE,
                                config_type=CONFIG_TYPE)
    upgrade.upgrade_user_config(user_config)
    user_config.save_changes()
except Exception as cfg_err:
    mlogger.debug('Can not read confing file at: %s | %s',
                    CONFIG_FILE, cfg_err)
    mlogger.debug('Using configs in memory...')
    user_config = verify_configs()



"""Revit events handler management."""
#pylint: disable=unused-argument
from pyrevit import HOST_APP
from pyrevit import EXEC_PARAMS, DB, UI
from pyrevit import framework
from pyrevit.coreutils.logger import get_logger


mlogger = get_logger(__name__)


class FuncAsEventHandler(UI.IExternalEventHandler):
    """Turns a function into an event handler."""
    __namespace__ = EXEC_PARAMS.exec_id

    def __init__(self, handler_func, purge=True):
        self.name = 'FuncAsEventHandler'
        self.handler_group_id = None
        self.handler_func = handler_func
        self.purge = purge

    def Execute(self, uiapp):
        if self.handler_func and self.handler_group_id:
            self.handler_func(self.handler_group_id)
            if self.purge:
                self.handler_group_id = self.handler_func = None

    def GetName(self):
        return self.name


def create_handler(handler_func, handler_args_type):
    return framework.EventHandler[handler_args_type](handler_func)


def add_handler(event_name, handler_func):
    event_handler = None

    if event_name == 'doc-changed':
        event_handler = \
            create_handler(handler_func, DB.Events.DocumentChangedEventArgs)
        HOST_APP.app.DocumentChanged += event_handler

    elif event_name == 'doc-closed':
        event_handler = \
            create_handler(handler_func, DB.Events.DocumentClosedEventArgs)
        HOST_APP.app.DocumentClosed += event_handler

    elif event_name == 'doc-opened':
        event_handler = \
            create_handler(handler_func, DB.Events.DocumentOpenedEventArgs)
        HOST_APP.app.DocumentOpened += event_handler

    elif event_name == 'view-activated':
        event_handler = \
            create_handler(handler_func, UI.Events.ViewActivatedEventArgs)
        HOST_APP.uiapp.ViewActivated += event_handler

    elif event_name == 'selection-changed':
        if HOST_APP.is_older_than(2023):
            mlogger.error("Not available in this Revit Version")
            return
        event_handler = \
            create_handler(handler_func, UI.Events.SelectionChangedEventArgs)
        HOST_APP.uiapp.SelectionChanged += event_handler

    return event_handler


def remove_handler(event_name, event_handler):
    if event_name == 'doc-changed':
        HOST_APP.app.DocumentChanged -= event_handler

    elif event_name == 'doc-closed':
        HOST_APP.app.DocumentClosed -= event_handler

    elif event_name == 'doc-opened':
        HOST_APP.app.DocumentOpened -= event_handler

    elif event_name == 'view-activated':
        HOST_APP.uiapp.ViewActivated -= event_handler

    elif event_name == 'selection-changed':
        if HOST_APP.is_older_than(2023):
            mlogger.error("Not available in this Revit Version")
            return
        HOST_APP.uiapp.SelectionChanged -= event_handler


def register_handler(event_name, handler_func, handler_group_id):
    if handler_group_id not in REGISTERED_HANDLERS:
        REGISTERED_HANDLERS[handler_group_id] = {}
    event_handler = add_handler(event_name, handler_func)
    if event_handler:
        REGISTERED_HANDLERS[handler_group_id][event_name] = event_handler


def unregister_handler(event_name, handler_func, handler_group_id):
    if handler_group_id in REGISTERED_HANDLERS \
            and event_name in REGISTERED_HANDLERS[handler_group_id]:
        event_handler = REGISTERED_HANDLERS[handler_group_id].pop(event_name)
        remove_handler(event_name, event_handler)


def unregister_exec_handlers(handler_group_id):
    if handler_group_id in REGISTERED_HANDLERS:
        for event_name, handler_func in \
                REGISTERED_HANDLERS[handler_group_id].items():
            unregister_handler(event_name, handler_func, handler_group_id)


def delayed_unregister_exec_handlers(handler_group_id):
    HANDLER_UNREGISTERER.handler_group_id = handler_group_id
    HANDLER_UNREGISTERER_EXTEVENT.Raise()


REGISTERED_HANDLERS = {}
HANDLER_UNREGISTERER = \
    FuncAsEventHandler(unregister_exec_handlers, purge=False)
HANDLER_UNREGISTERER_EXTEVENT = \
    UI.ExternalEvent.Create(HANDLER_UNREGISTERER)


def handle(*args): #pylint: disable=no-method-argument
    def decorator(function):
        def wrapper(*args, **kwargs):
            return function(*args, **kwargs)
        for event_name in args:
            register_handler(event_name, wrapper, EXEC_PARAMS.exec_id)
        return wrapper
    return decorator


def stop_events():
    if EXEC_PARAMS.exec_id in REGISTERED_HANDLERS:
        if HOST_APP.has_api_context:
            unregister_exec_handlers(EXEC_PARAMS.exec_id)
        else:
            # request underegister from external event
            delayed_unregister_exec_handlers(EXEC_PARAMS.exec_id)



class TimelineCollection(Animatable,ISealable,IAnimatable,IResource,IList,ICollection,IEnumerable,IList[Timeline],ICollection[Timeline],IEnumerable[Timeline]):
 """
 Represents a collection of System.Windows.Media.Animation.Timeline objects.

 

 TimelineCollection()

 TimelineCollection(capacity: int)

 TimelineCollection(collection: IEnumerable[Timeline])
 """
 def Add(self,value):
  """
  Add(self: TimelineCollection,value: Timeline)

   Inserts a new System.Windows.Media.Animation.Timeline object into the 

    System.Windows.Media.Animation.TimelineCollection.

  

  

   value: The object to add.
  """
  pass
 def Clear(self):
  """
  Clear(self: TimelineCollection)

   Removes all items from the System.Windows.Media.Animation.TimelineCollection.
  """
  pass
 def Clone(self):
  """
  Clone(self: TimelineCollection) -> TimelineCollection

  

   Creates a modifiable clone of this System.Windows.Media.Animation.TimelineCollection,making 

    deep copies of this object's values. When copying dependency properties,this method copies 

    resource references and data bindings (but they might no longer resolve) but not animations or 

    their current values.

  

   Returns: A modifiable clone of the current object. The cloned object's System.Windows.Freezable.IsFrozen 

    property will be false even if the source's System.Windows.Freezable.IsFrozen property was true.
  """
  pass
 def CloneCore(self,*args):
  """ CloneCore(self: TimelineCollection,source: Freezable) """
  pass
 def CloneCurrentValue(self):
  """
  CloneCurrentValue(self: TimelineCollection) -> TimelineCollection

  

   Creates a modifiable clone of this System.Windows.Media.Animation.TimelineCollection object,

    making deep copies of this object's current values. Resource references,data bindings,and 

    animations are not copied,but their current values are.

  

   Returns: A modifiable clone of the current object. The cloned object's System.Windows.Freezable.IsFrozen 

    property will be false even if the source's System.Windows.Freezable.IsFrozen property was true.
  """
  pass
 def CloneCurrentValueCore(self,*args):
  """ CloneCurrentValueCore(self: TimelineCollection,source: Freezable) """
  pass
 def Contains(self,value):
  """
  Contains(self: TimelineCollection,value: Timeline) -> bool

  

   Determines whether the System.Windows.Media.Animation.TimelineCollection contains the specified 

    System.Windows.Media.Animation.Timeline object.

  

  

   value: The object to locate.

   Returns: true if value is found; otherwise,false.
  """
  pass
 def CopyTo(self,array,index):
  """
  CopyTo(self: TimelineCollection,array: Array[Timeline],index: int)

   Copies the items of the System.Windows.Media.Animation.TimelineCollection to the passed Timeline 

    array,starting at the specified index position.

  

  

   array: The destination array.

   index: The zero-based index position where copying begins.
  """
  pass
 def CreateInstance(self,*args):
  """
  CreateInstance(self: Freezable) -> Freezable

  

   Initializes a new instance of the System.Windows.Freezable class.

   Returns: The new instance.
  """
  pass
 def CreateInstanceCore(self,*args):
  """ CreateInstanceCore(self: TimelineCollection) -> Freezable """
  pass
 def FreezeCore(self,*args):
  """ FreezeCore(self: TimelineCollection,isChecking: bool) -> bool """
  pass
 def GetAsFrozenCore(self,*args):
  """ GetAsFrozenCore(self: TimelineCollection,source: Freezable) """
  pass
 def GetCurrentValueAsFrozenCore(self,*args):
  """ GetCurrentValueAsFrozenCore(self: TimelineCollection,source: Freezable) """
  pass
 def GetEnumerator(self):
  """
  GetEnumerator(self: TimelineCollection) -> Enumerator

  

   Gets an enumerator that can iterate the members of the collection.

   Returns: An object that can iterate the members of the collection.
  """
  pass
 def IndexOf(self,value):
  """
  IndexOf(self: TimelineCollection,value: Timeline) -> int

  

   Gets the zero-based index position of a Timeline object in the 

    System.Windows.Media.Animation.TimelineCollection.

  

  

   value: The object to locate.

   Returns: The index position of value within this list.  If not found,-1 is returned.
  """
  pass
 def Insert(self,index,value):
  """
  Insert(self: TimelineCollection,index: int,value: Timeline)

   Inserts the specified Timeline object into this 

    System.Windows.Media.Animation.TimelineCollection at the specified index position.

  

  

   index: The zero-based index position into which value is inserted.

   value: The object to insert.
  """
  pass
 def OnChanged(self,*args):
  """
  OnChanged(self: Freezable)

   Called when the current System.Windows.Freezable object is modified.
  """
  pass
 def OnFreezablePropertyChanged(self,*args):
  """
  OnFreezablePropertyChanged(self: Freezable,oldValue: DependencyObject,newValue: DependencyObject,property: DependencyProperty)

   This member supports the Windows Presentation Foundation (WPF) infrastructure and is not 

    intended to be used directly from your code.

  

  

   oldValue: The previous value of the data member.

   newValue: The current value of the data member.

   property: The property that changed.

  OnFreezablePropertyChanged(self: Freezable,oldValue: DependencyObject,newValue: DependencyObject)

   Ensures that appropriate context pointers are established for a 

    System.Windows.DependencyObjectType data member that has just been set.

  

  

   oldValue: The previous value of the data member.

   newValue: The current value of the data member.
  """
  pass
 def OnPropertyChanged(self,*args):
  """
  OnPropertyChanged(self: Freezable,e: DependencyPropertyChangedEventArgs)

   Overrides the System.Windows.DependencyObject implementation of 

    System.Windows.DependencyObject.OnPropertyChanged(System.Windows.DependencyPropertyChangedEventAr

    gs) to also invoke any System.Windows.Freezable.Changed handlers in response to a changing 

    dependency property of type System.Windows.Freezable.

  

  

   e: Event data that contains information about which property changed,and its old and new values.
  """
  pass
 def ReadPreamble(self,*args):
  """
  ReadPreamble(self: Freezable)

   Ensures that the System.Windows.Freezable is being accessed from a valid thread. Inheritors of 

    System.Windows.Freezable must call this method at the beginning of any API that reads data 

    members that are not dependency properties.
  """
  pass
 def Remove(self,value):
  """
  Remove(self: TimelineCollection,value: Timeline) -> bool

  

   Removes the first occurrence of a specified System.Windows.Media.Animation.Timeline from this 

    System.Windows.Media.Animation.TimelineCollection.

  

  

   value: The object to remove.

   Returns: true if the operation was successful; otherwise,false.
  """
  pass
 def RemoveAt(self,index):
  """
  RemoveAt(self: TimelineCollection,index: int)

   Removes the System.Windows.Media.Animation.Timeline at the specified index position from this 

    System.Windows.Media.Animation.TimelineCollection.

  

  

   index: The zero-based index position of the item to remove.
  """
  pass
 def ShouldSerializeProperty(self,*args):
  """
  ShouldSerializeProperty(self: DependencyObject,dp: DependencyProperty) -> bool

  

   Returns a value that indicates whether serialization processes should serialize the value for 

    the provided dependency property.

  

  

   dp: The identifier for the dependency property that should be serialized.

   Returns: true if the dependency property that is supplied should be value-serialized; otherwise,false.
  """
  pass
 def WritePostscript(self,*args):
  """
  WritePostscript(self: Freezable)

   Raises the System.Windows.Freezable.Changed event for the System.Windows.Freezable and invokes 

    its System.Windows.Freezable.OnChanged method. Classes that derive from System.Windows.Freezable 

    should call this method at the end of any API that modifies class members that are not stored as 

    dependency properties.
  """
  pass
 def WritePreamble(self,*args):
  """
  WritePreamble(self: Freezable)

   Verifies that the System.Windows.Freezable is not frozen and that it is being accessed from a 

    valid threading context. System.Windows.Freezable inheritors should call this method at the 

    beginning of any API that writes to data members that are not dependency properties.
  """
  pass
 def __add__(self,*args):
  """ x.__add__(y) <==> x+y """
  pass
 def __contains__(self,*args):
  """
  __contains__(self: ICollection[Timeline],item: Timeline) -> bool

  __contains__(self: IList,value: object) -> bool

  

   Determines whether the System.Collections.IList contains a specific value.

  

   value: The object to locate in the System.Collections.IList.

   Returns: true if the System.Object is found in the System.Collections.IList; otherwise,false.
  """
  pass
 def __getitem__(self,*args):
  """ x.__getitem__(y) <==> x[y] """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __iter__(self,*args):
  """ __iter__(self: IEnumerable) -> object """
  pass
 def __len__(self,*args):
  """ x.__len__() <==> len(x) """
  pass
 @staticmethod
 def __new__(self,*__args):
  """
  __new__(cls: type)

  __new__(cls: type,capacity: int)

  __new__(cls: type,collection: IEnumerable[Timeline])
  """
  pass
 def __setitem__(self,*args):
  """ x.__setitem__(i,y) <==> x[i]= """
  pass
 Count=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the number of items contained in this System.Windows.Media.Animation.TimelineCollection.



Get: Count(self: TimelineCollection) -> int



"""


 Enumerator=None




class BitConverter(object):
 """ Converts base data types to an array of bytes,and an array of bytes to base data types. """
 @staticmethod
 def DoubleToInt64Bits(value):
  """
  DoubleToInt64Bits(value: float) -> Int64

  

   Converts the specified double-precision floating point number to a 64-bit signed integer.

  

   value: The number to convert.

   Returns: A 64-bit signed integer whose value is equivalent to value.
  """
  pass
 @staticmethod
 def GetBytes(value):
  """
  GetBytes(value: UInt32) -> Array[Byte]

  

   Returns the specified 32-bit unsigned integer value as an array of bytes.

  

   value: The number to convert.

   Returns: An array of bytes with length 4.

  GetBytes(value: UInt16) -> Array[Byte]

  

   Returns the specified 16-bit unsigned integer value as an array of bytes.

  

   value: The number to convert.

   Returns: An array of bytes with length 2.

  GetBytes(value: UInt64) -> Array[Byte]

  

   Returns the specified 64-bit unsigned integer value as an array of bytes.

  

   value: The number to convert.

   Returns: An array of bytes with length 8.

  GetBytes(value: float) -> Array[Byte]

  

   Returns the specified double-precision floating point value as an array of bytes.

  

   value: The number to convert.

   Returns: An array of bytes with length 8.

  GetBytes(value: Single) -> Array[Byte]

  

   Returns the specified single-precision floating point value as an array of bytes.

  

   value: The number to convert.

   Returns: An array of bytes with length 4.

  GetBytes(value: Char) -> Array[Byte]

  

   Returns the specified Unicode character value as an array of bytes.

  

   value: A character to convert.

   Returns: An array of bytes with length 2.

  GetBytes(value: bool) -> Array[Byte]

  

   Returns the specified Boolean value as an array of bytes.

  

   value: A Boolean value.

   Returns: An array of bytes with length 1.

  GetBytes(value: Int16) -> Array[Byte]

  

   Returns the specified 16-bit signed integer value as an array of bytes.

  

   value: The number to convert.

   Returns: An array of bytes with length 2.

  GetBytes(value: Int64) -> Array[Byte]

  

   Returns the specified 64-bit signed integer value as an array of bytes.

  

   value: The number to convert.

   Returns: An array of bytes with length 8.

  GetBytes(value: int) -> Array[Byte]

  

   Returns the specified 32-bit signed integer value as an array of bytes.

  

   value: The number to convert.

   Returns: An array of bytes with length 4.
  """
  pass
 @staticmethod
 def Int64BitsToDouble(value):
  """
  Int64BitsToDouble(value: Int64) -> float

  

   Converts the specified 64-bit signed integer to a double-precision floating point number.

  

   value: The number to convert.

   Returns: A double-precision floating point number whose value is equivalent to value.
  """
  pass
 @staticmethod
 def ToBoolean(value,startIndex):
  """
  ToBoolean(value: Array[Byte],startIndex: int) -> bool

  

   Returns a Boolean value converted from one byte at a specified position in a byte array.

  

   value: An array of bytes.

   startIndex: The starting position within value.

   Returns: true if the byte at startIndex in value is nonzero; otherwise,false.
  """
  pass
 @staticmethod
 def ToChar(value,startIndex):
  """
  ToChar(value: Array[Byte],startIndex: int) -> Char

  

   Returns a Unicode character converted from two bytes at a specified position in a byte array.

  

   value: An array.

   startIndex: The starting position within value.

   Returns: A character formed by two bytes beginning at startIndex.
  """
  pass
 @staticmethod
 def ToDouble(value,startIndex):
  """
  ToDouble(value: Array[Byte],startIndex: int) -> float

  

   Returns a double-precision floating point number converted from eight bytes at a specified 

    position in a byte array.

  

  

   value: An array of bytes.

   startIndex: The starting position within value.

   Returns: A double precision floating point number formed by eight bytes beginning at startIndex.
  """
  pass
 @staticmethod
 def ToInt16(value,startIndex):
  """
  ToInt16(value: Array[Byte],startIndex: int) -> Int16

  

   Returns a 16-bit signed integer converted from two bytes at a specified position in a byte array.

  

   value: An array of bytes.

   startIndex: The starting position within value.

   Returns: A 16-bit signed integer formed by two bytes beginning at startIndex.
  """
  pass
 @staticmethod
 def ToInt32(value,startIndex):
  """
  ToInt32(value: Array[Byte],startIndex: int) -> int

  

   Returns a 32-bit signed integer converted from four bytes at a specified position in a byte 

    array.

  

  

   value: An array of bytes.

   startIndex: The starting position within value.

   Returns: A 32-bit signed integer formed by four bytes beginning at startIndex.
  """
  pass
 @staticmethod
 def ToInt64(value,startIndex):
  """
  ToInt64(value: Array[Byte],startIndex: int) -> Int64

  

   Returns a 64-bit signed integer converted from eight bytes at a specified position in a byte 

    array.

  

  

   value: An array of bytes.

   startIndex: The starting position within value.

   Returns: A 64-bit signed integer formed by eight bytes beginning at startIndex.
  """
  pass
 @staticmethod
 def ToSingle(value,startIndex):
  """
  ToSingle(value: Array[Byte],startIndex: int) -> Single

  

   Returns a single-precision floating point number converted from four bytes at a specified 

    position in a byte array.

  

  

   value: An array of bytes.

   startIndex: The starting position within value.

   Returns: A single-precision floating point number formed by four bytes beginning at startIndex.
  """
  pass
 @staticmethod
 def ToString(value=None,startIndex=None,length=None):
  """
  ToString(value: Array[Byte],startIndex: int) -> str

  

   Converts the numeric value of each element of a specified subarray of bytes to its equivalent 

    hexadecimal string representation.

  

  

   value: An array of bytes.

   startIndex: The starting position within value.

   Returns: A string of hexadecimal pairs separated by hyphens,where each pair represents the corresponding 

    element in a subarray of value; for example,"7F-2C-4A-00".

  

  ToString(value: Array[Byte]) -> str

  

   Converts the numeric value of each element of a specified array of bytes to its equivalent 

    hexadecimal string representation.

  

  

   value: An array of bytes.

   Returns: A string of hexadecimal pairs separated by hyphens,where each pair represents the corresponding 

    element in value; for example,"7F-2C-4A-00".

  

  ToString(value: Array[Byte],startIndex: int,length: int) -> str

  

   Converts the numeric value of each element of a specified subarray of bytes to its equivalent 

    hexadecimal string representation.

  

  

   value: An array of bytes.

   startIndex: The starting position within value.

   length: The number of array elements in value to convert.

   Returns: A string of hexadecimal pairs separated by hyphens,where each pair represents the corresponding 

    element in a subarray of value; for example,"7F-2C-4A-00".
  """
  pass
 @staticmethod
 def ToUInt16(value,startIndex):
  """
  ToUInt16(value: Array[Byte],startIndex: int) -> UInt16

  

   Returns a 16-bit unsigned integer converted from two bytes at a specified position in a byte 

    array.

  

  

   value: The array of bytes.

   startIndex: The starting position within value.

   Returns: A 16-bit unsigned integer formed by two bytes beginning at startIndex.
  """
  pass
 @staticmethod
 def ToUInt32(value,startIndex):
  """
  ToUInt32(value: Array[Byte],startIndex: int) -> UInt32

  

   Returns a 32-bit unsigned integer converted from four bytes at a specified position in a byte 

    array.

  

  

   value: An array of bytes.

   startIndex: The starting position within value.

   Returns: A 32-bit unsigned integer formed by four bytes beginning at startIndex.
  """
  pass
 @staticmethod
 def ToUInt64(value,startIndex):
  """
  ToUInt64(value: Array[Byte],startIndex: int) -> UInt64

  

   Returns a 64-bit unsigned integer converted from eight bytes at a specified position in a byte 

    array.

  

  

   value: An array of bytes.

   startIndex: The starting position within value.

   Returns: A 64-bit unsigned integer formed by the eight bytes beginning at startIndex.
  """
  pass
 IsLittleEndian=True
 __all__=[
  'DoubleToInt64Bits',
  'GetBytes',
  'Int64BitsToDouble',
  'IsLittleEndian',
  'ToBoolean',
  'ToChar',
  'ToDouble',
  'ToInt16',
  'ToInt32',
  'ToInt64',
  'ToSingle',
  'ToString',
  'ToUInt16',
  'ToUInt32',
  'ToUInt64',
 ]




"""Base module ofr parsing extensions."""
import os
import os.path as op

from pyrevit.coreutils import get_all_subclasses
from pyrevit.coreutils.logger import get_logger


#pylint: disable=W0703,C0302,C0103
mlogger = get_logger(__name__)


def _get_discovered_comps(comp_path, cmp_types_list):
    discovered_cmps = []
    mlogger.debug('Testing _get_component(s) on: %s ', comp_path)
    # comp_path might be a file or a dir,
    # but its name should not start with . or _:
    for cmp_type in cmp_types_list:
        mlogger.debug('Testing sub_directory %s for %s', comp_path, cmp_type)
        # if cmp_class can be created for this sub-dir, the add to list
        if cmp_type.matches(comp_path):
            component = cmp_type(cmp_path=comp_path)
            discovered_cmps.append(component)
            mlogger.debug('Successfuly created component: %s from: %s',
                          component, comp_path)

    return discovered_cmps


def _create_subcomponents(search_dir,
                          cmp_types_list,
                          create_from_search_dir=False):
    """Returns the objects in search_dir of the types in cmp_types_list.

    Arguments:
        search_dir (str): directory to parse
        cmp_types_list: This methods checks the subfolders in search_dir
            against the _get_component types provided in this list.
        create_from_search_dir (bool, optional): whether to create the _get_component objects, default to False

    Examples:
        ```python
        _create_subcomponents(search_dir,
                              [LinkButton, PushButton, or ToggleButton])
        ```
        this method creates LinkButton, PushButton, or ToggleButton for
        the parsed sub-directories under search_dir with matching .type_id
        identifiers in their names. (e.g. "folder.LINK_BUTTON_POSTFIX")

    Returns:
        list of created classes of types provided in cmp_types_list
    """
    sub_cmp_list = []

    if not create_from_search_dir:
        mlogger.debug('Searching directory: %s for components of type: %s',
                      search_dir, cmp_types_list)
        for file_or_dir in os.listdir(search_dir):
            full_path = op.join(search_dir, file_or_dir)
            if not file_or_dir.startswith(('.', '_')):
                sub_cmp_list.extend(_get_discovered_comps(full_path,
                                                          cmp_types_list))
            else:
                mlogger.debug('Skipping _get_component. '
                              'Name can not start with . or _: %s', full_path)
    else:
        sub_cmp_list.extend(_get_discovered_comps(search_dir,
                                                  cmp_types_list))

    return sub_cmp_list


def _get_subcomponents_classes(parent_classes):
    """Find available subcomponents for given parent types."""
    return [x for x in get_all_subclasses(parent_classes) if x.type_id]


def _parse_for_components(component):
    """Recursively parses the component directory for allowed components type.

    This method uses get_all_subclasses() to get a list of all subclasses
    of _get_component.allowed_sub_cmps type.
    This ensures that if any new type of component_class is added later,
    this method does not need to be updated as the new sub-class will be
    listed by .__subclasses__() method of the parent class and this method
    will check the directory for its .type_id.
    """
    for new_cmp in _create_subcomponents(
            component.directory,
            _get_subcomponents_classes(component.allowed_sub_cmps)):
        # add the successfulyl created _get_component to the
        # parent _get_component
        component.add_component(new_cmp)
        if new_cmp.is_container:
            # Recursive part: parse each sub-_get_component
            # for its allowed sub-sub-components.
            _parse_for_components(new_cmp)


def parse_comp_dir(comp_path, comp_class):
    return _create_subcomponents(
        comp_path,
        _get_subcomponents_classes([comp_class]),
        create_from_search_dir=True
        )


def get_parsed_extension(extension):
    """Creates and adds the extensions components to the package.

    Each package object is the root to a tree of components that exists
    under that package. (e.g. tabs, buttons, ...) sub components of package
    can be accessed by iterating the _get_component.
    See _basecomponents for types.
    """
    _parse_for_components(extension)
    return extension


def parse_dir_for_ext_type(root_dir, parent_cmp_type):
    """Return the objects of type parent_cmp_type of the extensions in root_dir.

    The package objects won't be parsed at this level.
    This is useful for collecting basic info on an extension type
    for cache cheching or updating extensions using their directory paths.

    Args:
        root_dir (str): directory to parse
        parent_cmp_type (type): type of objects to return
    """
    # making sure the provided directory exists.
    # This is mainly for the user defined package directories
    if not op.exists(root_dir):
        mlogger.debug('Extension search directory does not exist: %s', root_dir)
        return []

    # try creating extensions in given directory
    ext_data_list = []

    mlogger.debug('Parsing directory for extensions of type: %s',
                  parent_cmp_type)
    for ext_data in _create_subcomponents(root_dir, [parent_cmp_type]):
        mlogger.debug('Extension directory found: %s', ext_data)
        ext_data_list.append(ext_data)

    return ext_data_list



"""
Use these exceptions to `try` against specific Rpw Exceptions.

>>> from rpw.exceptions import RpwWrongStorageType
>>> try:
...     element.parameters['Height'].value = 'String'
... except RpwWrongStorageType:
...     print('Height Parameter cannot be a string')
...     raise

This module also provides easy access to the ``Autodesk.Revit.Exceptions``
namespaces:

>>> from rpw.exceptions import RevitExceptions
>>> try:
...     doc.Delete(ElementId)
... except RevitExceptions.InvalidObjectException:
...     print('This element is no longer valid ')

"""  #

from rpw.utils.logger import logger
# Added on 1.7.4
# Since adding Autodesk.Revit, it became impossible to run any non-revit
# tools such as forms.os_forms, etc using run_forms.BaseException
# This needs clean up
from rpw.utils.sphinx_compat import MockObject
try:
    from Autodesk.Revit import Exceptions as RevitExceptions
except ImportError:
    RevitExceptions = MockObject()


class RpwException(Exception):
    """ Revit Python Wrapper Base Exception """


class RpwTypeError(TypeError):
    """ Revit Python Wrapper Type Exception """
    def __init__(self, type_expected, type_received='not reported'):
        msg = 'expected [{}], got [{}]'.format(type_expected, type_received)
        super(RpwTypeError, self).__init__(msg)


class RpwValueError(ValueError):
    """ Revit Python Wrapper Value Error Exception """
    def __init__(self, value_expected, value_received='not reported'):
        msg = 'expected [{}], got [{}]'.format(value_expected, value_received)
        super(RpwValueError, self).__init__(msg)


class RpwParameterNotFound(RpwException, KeyError):
    """ Revit Python Wrapper Parameter Error """
    def __init__(self, element, param_name):
        msg = 'parameter not found [element:{}]:[param_name:{}]'.format(
                                                        element.Id, param_name)
        super(RpwParameterNotFound, self).__init__(msg)


class RpwWrongStorageType(RpwException, TypeError):
    """ Wrong Storage Type """
    def __init__(self, storage_type, value):
        msg = 'Wrong Storage Type: [{}]:[{}:{}]'.format(storage_type,
                                                        type(value), value)
        super(RpwWrongStorageType, self).__init__(msg)


class RpwCoerceError(RpwException, ValueError):
    """ Coerce Error """
    def __init__(self, value, target_type):
        msg = 'Could not cast value:{} to target_type:{}'.format(value,
                                                                 target_type)
        super(RpwCoerceError, self).__init__(msg)



class PropertyManager(BindingManagerBase):
 """
 Maintains a System.Windows.Forms.Binding between an object's property and a data-bound control property.

 

 PropertyManager()
 """
 def AddNew(self):
  """ AddNew(self: PropertyManager) """
  pass
 def CancelCurrentEdit(self):
  """ CancelCurrentEdit(self: PropertyManager) """
  pass
 def EndCurrentEdit(self):
  """ EndCurrentEdit(self: PropertyManager) """
  pass
 def GetItemProperties(self):
  """
  GetItemProperties(self: BindingManagerBase,listType: Type,offset: int,dataSources: ArrayList,listAccessors: ArrayList) -> PropertyDescriptorCollection

  

   Gets the list of properties of the items managed by this System.Windows.Forms.BindingManagerBase.

  

   listType: The System.Type of the bound list.

   offset: A counter used to recursively call the method.

   dataSources: An System.Collections.ArrayList containing the data sources.

   listAccessors: An System.Collections.ArrayList containing the table's bound properties.

   Returns: A System.ComponentModel.PropertyDescriptorCollection that represents the property descriptors 

    for the binding.

  

  GetItemProperties(self: BindingManagerBase,dataSources: ArrayList,listAccessors: ArrayList) -> PropertyDescriptorCollection

  

   Gets the collection of property descriptors for the binding using the specified 

    System.Collections.ArrayList.

  

  

   dataSources: An System.Collections.ArrayList containing the data sources.

   listAccessors: An System.Collections.ArrayList containing the table's bound properties.

   Returns: A System.ComponentModel.PropertyDescriptorCollection that represents the property descriptors 

    for the binding.
  """
  pass
 def RemoveAt(self,index):
  """
  RemoveAt(self: PropertyManager,index: int)

   index: The index of the row to delete.
  """
  pass
 def ResumeBinding(self):
  """ ResumeBinding(self: PropertyManager) """
  pass
 def SuspendBinding(self):
  """
  SuspendBinding(self: PropertyManager)

   Suspends the data binding between a data source and a data-bound property.
  """
  pass
 Count=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Get: Count(self: PropertyManager) -> int



"""

 Current=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the object to which the data-bound property belongs.



Get: Current(self: PropertyManager) -> object



"""

 Position=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Get: Position(self: PropertyManager) -> int



Set: Position(self: PropertyManager)=value

"""


 onCurrentChangedHandler=None
 onPositionChangedHandler=None




class ComEventInterfaceAttribute(Attribute,_Attribute):
 """
 Identifies the source interface and the class that implements the methods of the event interface that is generated when a coclass is imported from a COM type library.

 

 ComEventInterfaceAttribute(SourceInterface: Type,EventProvider: Type)
 """
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 @staticmethod
 def __new__(self,SourceInterface,EventProvider):
  """ __new__(cls: type,SourceInterface: Type,EventProvider: Type) """
  pass
 EventProvider=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the class that implements the methods of the event interface.



Get: EventProvider(self: ComEventInterfaceAttribute) -> Type



"""

 SourceInterface=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the original source interface from the type library.



Get: SourceInterface(self: ComEventInterfaceAttribute) -> Type



"""





class TypedReference(object):
 """ Describes objects that contain both a managed pointer to a location and a runtime representation of the type that may be stored at that location. """
 def Equals(self,o):
  """
  Equals(self: TypedReference,o: object) -> bool

  

   Checks if this object is equal to the specified object.

  

   o: The object with which to compare the current object.

   Returns: true if this object is equal to the specified object; otherwise,false.
  """
  pass
 def GetHashCode(self):
  """
  GetHashCode(self: TypedReference) -> int

  

   Returns the hash code of this object.

   Returns: The hash code of this object.
  """
  pass
 @staticmethod
 def GetTargetType(value):
  """
  GetTargetType(value: TypedReference) -> Type

  

   Returns the type of the target of the specified TypedReference.

  

   value: The value whose target's type is to be returned.

   Returns: The type of the target of the specified TypedReference.
  """
  pass
 @staticmethod
 def MakeTypedReference(target,flds):
  """
  MakeTypedReference(target: object,flds: Array[FieldInfo]) -> TypedReference

  

   Makes a TypedReference for a field identified by a specified object and list of field 

    descriptions.

  

  

   target: An object that contains the field described by the first element of flds.

   flds: A list of field descriptions where each element describes a field that contains the field 

    described by the succeeding element. Each described field must be a value type. The field 

    descriptions must be RuntimeFieldInfo objects supplied by the type system.

  

   Returns: A System.TypedReference for the field described by the last element of flds.
  """
  pass
 @staticmethod
 def SetTypedReference(target,value):
  """
  SetTypedReference(target: TypedReference,value: object)

   Converts the specified value to a TypedReference. This method is not supported.

  

   target: The target of the conversion.

   value: The value to be converted.
  """
  pass
 @staticmethod
 def TargetTypeToken(value):
  """
  TargetTypeToken(value: TypedReference) -> RuntimeTypeHandle

  

   Returns the internal metadata type handle for the specified TypedReference.

  

   value: The TypedReference for which the type handle is requested.

   Returns: The internal metadata type handle for the specified TypedReference.
  """
  pass
 @staticmethod
 def ToObject(value):
  """
  ToObject(value: TypedReference) -> object

  

   Converts the specified TypedReference to an Object.

  

   value: The TypedReference to be converted.

   Returns: An System.Object converted from a TypedReference.
  """
  pass
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==y """
  pass
 def __ne__(self,*args):
  pass



class InputLanguage(object):
 """ Provides methods and fields to manage the input language. This class cannot be inherited. """
 def Equals(self,value):
  """
  Equals(self: InputLanguage,value: object) -> bool

  

   Specifies whether two input languages are equal.

  

   value: The language to test for equality.

   Returns: true if the two languages are equal; otherwise,false.
  """
  pass
 @staticmethod
 def FromCulture(culture):
  """
  FromCulture(culture: CultureInfo) -> InputLanguage

  

   Returns the input language associated with the specified culture.

  

   culture: The System.Globalization.CultureInfo that specifies the culture to convert from.

   Returns: An System.Windows.Forms.InputLanguage that represents the previously selected input language.
  """
  pass
 def GetHashCode(self):
  """
  GetHashCode(self: InputLanguage) -> int

  

   Returns the hash code for this input language.

   Returns: The hash code for this input language.
  """
  pass
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==y """
  pass
 def __ne__(self,*args):
  pass
 Culture=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the culture of the current input language.



Get: Culture(self: InputLanguage) -> CultureInfo



"""

 Handle=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the handle for the input language.



Get: Handle(self: InputLanguage) -> IntPtr



"""

 LayoutName=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the name of the current keyboard layout as it appears in the regional settings of the operating system on the computer.



Get: LayoutName(self: InputLanguage) -> str



"""


 CurrentInputLanguage=None
 DefaultInputLanguage=None
 InstalledInputLanguages=None




class HttpListenerBasicIdentity(GenericIdentity,IIdentity):
 """
 Holds the user name and password from a basic authentication request.

 

 HttpListenerBasicIdentity(username: str,password: str)
 """
 def CreateClaim(self,*args):
  """ CreateClaim(self: ClaimsIdentity,reader: BinaryReader) -> Claim """
  pass
 def GetObjectData(self,*args):
  """ GetObjectData(self: ClaimsIdentity,info: SerializationInfo,context: StreamingContext) """
  pass
 def WriteTo(self,writer):
  """ WriteTo(self: ClaimsIdentity,writer: BinaryWriter,userData: Array[Byte]) """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 @staticmethod
 def __new__(self,username,password):
  """ __new__(cls: type,username: str,password: str) """
  pass
 CustomSerializationData=property(lambda self: object(),lambda self,v: None,lambda self: None)

 Password=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Indicates the password from a basic authentication attempt.



Get: Password(self: HttpListenerBasicIdentity) -> str



"""





class NativeWindow(MarshalByRefObject,IWin32Window):
 """
 Provides a low-level encapsulation of a window handle and a window procedure.

 

 NativeWindow()
 """
 def AssignHandle(self,handle):
  """
  AssignHandle(self: NativeWindow,handle: IntPtr)

   Assigns a handle to this window.

  

   handle: The handle to assign to this window.
  """
  pass
 def CreateHandle(self,cp):
  """
  CreateHandle(self: NativeWindow,cp: CreateParams)

   Creates a window and its handle with the specified creation parameters.

  

   cp: A System.Windows.Forms.CreateParams that specifies the creation parameters for this window.
  """
  pass
 def DefWndProc(self,m):
  """
  DefWndProc(self: NativeWindow,m: Message) -> Message

  

   Invokes the default window procedure associated with this window.

  

   m: The message that is currently being processed.
  """
  pass
 def DestroyHandle(self):
  """
  DestroyHandle(self: NativeWindow)

   Destroys the window and its handle.
  """
  pass
 @staticmethod
 def FromHandle(handle):
  """
  FromHandle(handle: IntPtr) -> NativeWindow

  

   Retrieves the window associated with the specified handle.

  

   handle: A handle to a window.

   Returns: The System.Windows.Forms.NativeWindow associated with the specified handle. This method returns 

    null when the handle does not have an associated window.
  """
  pass
 def MemberwiseClone(self,*args):
  """
  MemberwiseClone(self: MarshalByRefObject,cloneIdentity: bool) -> MarshalByRefObject

  

   Creates a shallow copy of the current System.MarshalByRefObject object.

  

   cloneIdentity: false to delete the current System.MarshalByRefObject object's identity,which will cause the 

    object to be assigned a new identity when it is marshaled across a remoting boundary. A value of 

    false is usually appropriate. true to copy the current System.MarshalByRefObject object's 

    identity to its clone,which will cause remoting client calls to be routed to the remote server 

    object.

  

   Returns: A shallow copy of the current System.MarshalByRefObject object.

  MemberwiseClone(self: object) -> object

  

   Creates a shallow copy of the current System.Object.

   Returns: A shallow copy of the current System.Object.
  """
  pass
 def OnHandleChange(self,*args):
  """
  OnHandleChange(self: NativeWindow)

   Specifies a notification method that is called when the handle for a window is changed.
  """
  pass
 def OnThreadException(self,*args):
  """
  OnThreadException(self: NativeWindow,e: Exception)

   When overridden in a derived class,manages an unhandled thread exception.

  

   e: An System.Exception that specifies the unhandled thread exception.
  """
  pass
 def ReleaseHandle(self):
  """
  ReleaseHandle(self: NativeWindow)

   Releases the handle associated with this window.
  """
  pass
 def WndProc(self,*args):
  """
  WndProc(self: NativeWindow,m: Message) -> Message

  

   Invokes the default window procedure associated with this window.

  

   m: A System.Windows.Forms.Message that is associated with the current Windows message.
  """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 Handle=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the handle for this window.



Get: Handle(self: NativeWindow) -> IntPtr



"""





class Authorization(object):
 """
 Contains an authentication message for an Internet server.

 

 Authorization(token: str)

 Authorization(token: str,finished: bool)

 Authorization(token: str,finished: bool,connectionGroupId: str)
 """
 @staticmethod
 def __new__(self,token,finished=None,connectionGroupId=None):
  """
  __new__(cls: type,token: str)

  __new__(cls: type,token: str,finished: bool)

  __new__(cls: type,token: str,finished: bool,connectionGroupId: str)
  """
  pass
 Complete=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the completion status of the authorization.



Get: Complete(self: Authorization) -> bool



"""

 ConnectionGroupId=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a unique identifier for user-specific connections.



Get: ConnectionGroupId(self: Authorization) -> str



"""

 Message=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the message returned to the server in response to an authentication challenge.



Get: Message(self: Authorization) -> str



"""

 MutuallyAuthenticated=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets a System.Boolean value that indicates whether mutual authentication occurred.



Get: MutuallyAuthenticated(self: Authorization) -> bool



Set: MutuallyAuthenticated(self: Authorization)=value

"""

 ProtectionRealm=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the prefix for Uniform Resource Identifiers (URIs) that can be authenticated with the System.Net.Authorization.Message property.



Get: ProtectionRealm(self: Authorization) -> Array[str]



Set: ProtectionRealm(self: Authorization)=value

"""





class DataGridViewDataErrorEventArgs(DataGridViewCellCancelEventArgs):
 """
 Provides data for the System.Windows.Forms.DataGridView.DataError event.

 

 DataGridViewDataErrorEventArgs(exception: Exception,columnIndex: int,rowIndex: int,context: DataGridViewDataErrorContexts)
 """
 @staticmethod
 def __new__(self,exception,columnIndex,rowIndex,context):
  """ __new__(cls: type,exception: Exception,columnIndex: int,rowIndex: int,context: DataGridViewDataErrorContexts) """
  pass
 Context=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets details about the state of the System.Windows.Forms.DataGridView when the error occurred.



Get: Context(self: DataGridViewDataErrorEventArgs) -> DataGridViewDataErrorContexts



"""

 Exception=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the exception that represents the error.



Get: Exception(self: DataGridViewDataErrorEventArgs) -> Exception



"""

 ThrowException=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets a value indicating whether to throw the exception after the System.Windows.Forms.DataGridViewDataErrorEventHandler delegate is finished with it.



Get: ThrowException(self: DataGridViewDataErrorEventArgs) -> bool



Set: ThrowException(self: DataGridViewDataErrorEventArgs)=value

"""





class FormClosingEventArgs(CancelEventArgs):
 """
 Provides data for the System.Windows.Forms.Form.FormClosing event.

 

 FormClosingEventArgs(closeReason: CloseReason,cancel: bool)
 """
 @staticmethod
 def __new__(self,closeReason,cancel):
  """ __new__(cls: type,closeReason: CloseReason,cancel: bool) """
  pass
 CloseReason=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value that indicates why the form is being closed.



Get: CloseReason(self: FormClosingEventArgs) -> CloseReason



"""





class DataObject(object,IDataObject,IDataObject):
 """
 Implements a basic data transfer mechanism.

 

 DataObject()

 DataObject(data: object)

 DataObject(format: str,data: object)
 """
 def ContainsAudio(self):
  """
  ContainsAudio(self: DataObject) -> bool

  

   Indicates whether the data object contains data in the 

    System.Windows.Forms.DataFormats.WaveAudio format.

  

   Returns: true if the data object contains audio data; otherwise,false.
  """
  pass
 def ContainsFileDropList(self):
  """
  ContainsFileDropList(self: DataObject) -> bool

  

   Indicates whether the data object contains data that is in the 

    System.Windows.Forms.DataFormats.FileDrop format or can be converted to that format.

  

   Returns: true if the data object contains a file drop list; otherwise,false.
  """
  pass
 def ContainsImage(self):
  """
  ContainsImage(self: DataObject) -> bool

  

   Indicates whether the data object contains data that is in the 

    System.Windows.Forms.DataFormats.Bitmap format or can be converted to that format.

  

   Returns: true if the data object contains image data; otherwise,false.
  """
  pass
 def ContainsText(self,format=None):
  """
  ContainsText(self: DataObject,format: TextDataFormat) -> bool

  

   Indicates whether the data object contains text data in the format indicated by the specified 

    System.Windows.Forms.TextDataFormat value.

  

  

   format: One of the System.Windows.Forms.TextDataFormat values.

   Returns: true if the data object contains text data in the specified format; otherwise,false.

  ContainsText(self: DataObject) -> bool

  

   Indicates whether the data object contains data in the 

    System.Windows.Forms.TextDataFormat.UnicodeText format.

  

   Returns: true if the data object contains text data; otherwise,false.
  """
  pass
 def GetAudioStream(self):
  """
  GetAudioStream(self: DataObject) -> Stream

  

   Retrieves an audio stream from the data object.

   Returns: A System.IO.Stream containing audio data or null if the data object does not contain any data in 

    the System.Windows.Forms.DataFormats.WaveAudio format.
  """
  pass
 def GetData(self,format,autoConvert=None):
  """
  GetData(self: DataObject,format: Type) -> object

  

   Returns the data associated with the specified class type format.

  

   format: A System.Type representing the format of the data to retrieve.

   Returns: The data associated with the specified format,or null.

  GetData(self: DataObject,format: str) -> object

  

   Returns the data associated with the specified data format.

  

   format: The format of the data to retrieve. See System.Windows.Forms.DataFormats for predefined formats.

   Returns: The data associated with the specified format,or null.

  GetData(self: DataObject,format: str,autoConvert: bool) -> object

  

   Returns the data associated with the specified data format,using an automated conversion 

    parameter to determine whether to convert the data to the format.

  

  

   format: The format of the data to retrieve. See System.Windows.Forms.DataFormats for predefined formats.

   autoConvert: true to the convert data to the specified format; otherwise,false.

   Returns: The data associated with the specified format,or null.
  """
  pass
 def GetDataPresent(self,format,autoConvert=None):
  """
  GetDataPresent(self: DataObject,format: str) -> bool

  

   Determines whether data stored in this System.Windows.Forms.DataObject is associated with,or 

    can be converted to,the specified format.

  

  

   format: The format to check for. See System.Windows.Forms.DataFormats for predefined formats.

   Returns: true if data stored in this System.Windows.Forms.DataObject is associated with,or can be 

    converted to,the specified format; otherwise,false.

  

  GetDataPresent(self: DataObject,format: str,autoConvert: bool) -> bool

  

   Determines whether this System.Windows.Forms.DataObject contains data in the specified format 

    or,optionally,contains data that can be converted to the specified format.

  

  

   format: The format to check for. See System.Windows.Forms.DataFormats for predefined formats.

   autoConvert: true to determine whether data stored in this System.Windows.Forms.DataObject can be converted 

    to the specified format; false to check whether the data is in the specified format.

  

   Returns: true if the data is in,or can be converted to,the specified format; otherwise,false.

  GetDataPresent(self: DataObject,format: Type) -> bool

  

   Determines whether data stored in this System.Windows.Forms.DataObject is associated with,or 

    can be converted to,the specified format.

  

  

   format: A System.Type representing the format to check for.

   Returns: true if data stored in this System.Windows.Forms.DataObject is associated with,or can be 

    converted to,the specified format; otherwise,false.
  """
  pass
 def GetFileDropList(self):
  """
  GetFileDropList(self: DataObject) -> StringCollection

  

   Retrieves a collection of file names from the data object.

   Returns: A System.Collections.Specialized.StringCollection containing file names or null if the data 

    object does not contain any data that is in the System.Windows.Forms.DataFormats.FileDrop format 

    or can be converted to that format.
  """
  pass
 def GetFormats(self,autoConvert=None):
  """
  GetFormats(self: DataObject) -> Array[str]

  

   Returns a list of all formats that data stored in this System.Windows.Forms.DataObject is 

    associated with or can be converted to.

  

   Returns: An array of type System.String,containing a list of all formats that are supported by the data 

    stored in this object.

  

  GetFormats(self: DataObject,autoConvert: bool) -> Array[str]

  

   Returns a list of all formats that data stored in this System.Windows.Forms.DataObject is 

    associated with or can be converted to,using an automatic conversion parameter to determine 

    whether to retrieve only native data formats or all formats that the data can be converted to.

  

  

   autoConvert: true to retrieve all formats that data stored in this System.Windows.Forms.DataObject is 

    associated with,or can be converted to; false to retrieve only native data formats.

  

   Returns: An array of type System.String,containing a list of all formats that are supported by the data 

    stored in this object.
  """
  pass
 def GetImage(self):
  """
  GetImage(self: DataObject) -> Image

  

   Retrieves an image from the data object.

   Returns: An System.Drawing.Image representing the image data in the data object or null if the data 

    object does not contain any data that is in the System.Windows.Forms.DataFormats.Bitmap format 

    or can be converted to that format.
  """
  pass
 def GetText(self,format=None):
  """
  GetText(self: DataObject,format: TextDataFormat) -> str

  

   Retrieves text data from the data object in the format indicated by the specified 

    System.Windows.Forms.TextDataFormat value.

  

  

   format: One of the System.Windows.Forms.TextDataFormat values.

   Returns: The text data in the data object or System.String.Empty if the data object does not contain data 

    in the specified format.

  

  GetText(self: DataObject) -> str

  

   Retrieves text data from the data object in the System.Windows.Forms.TextDataFormat.UnicodeText 

    format.

  

   Returns: The text data in the data object or System.String.Empty if the data object does not contain data 

    in the System.Windows.Forms.TextDataFormat.UnicodeText format.
  """
  pass
 def SetAudio(self,*__args):
  """
  SetAudio(self: DataObject,audioStream: Stream)

   Adds a System.IO.Stream to the data object in the System.Windows.Forms.DataFormats.WaveAudio 

    format.

  

  

   audioStream: A System.IO.Stream containing the audio data.

  SetAudio(self: DataObject,audioBytes: Array[Byte])

   Adds a System.Byte array to the data object in the System.Windows.Forms.DataFormats.WaveAudio 

    format after converting it to a System.IO.Stream.

  

  

   audioBytes: A System.Byte array containing the audio data.
  """
  pass
 def SetData(self,*__args):
  """
  SetData(self: DataObject,format: Type,data: object)

   Adds the specified object to the System.Windows.Forms.DataObject using the specified type as the 

    format.

  

  

   format: A System.Type representing the format associated with the data.

   data: The data to store.

  SetData(self: DataObject,data: object)

   Adds the specified object to the System.Windows.Forms.DataObject using the object type as the 

    data format.

  

  

   data: The data to store.

  SetData(self: DataObject,format: str,autoConvert: bool,data: object)

   Adds the specified object to the System.Windows.Forms.DataObject using the specified format and 

    indicating whether the data can be converted to another format.

  

  

   format: The format associated with the data. See System.Windows.Forms.DataFormats for predefined formats.

   autoConvert: true to allow the data to be converted to another format; otherwise,false.

   data: The data to store.

  SetData(self: DataObject,format: str,data: object)

   Adds the specified object to the System.Windows.Forms.DataObject using the specified format.

  

   format: The format associated with the data. See System.Windows.Forms.DataFormats for predefined formats.

   data: The data to store.
  """
  pass
 def SetFileDropList(self,filePaths):
  """
  SetFileDropList(self: DataObject,filePaths: StringCollection)

   Adds a collection of file names to the data object in the 

    System.Windows.Forms.DataFormats.FileDrop format.

  

  

   filePaths: A System.Collections.Specialized.StringCollection containing the file names.
  """
  pass
 def SetImage(self,image):
  """
  SetImage(self: DataObject,image: Image)

   Adds an System.Drawing.Image to the data object in the System.Windows.Forms.DataFormats.Bitmap 

    format.

  

  

   image: The System.Drawing.Image to add to the data object.
  """
  pass
 def SetText(self,textData,format=None):
  """
  SetText(self: DataObject,textData: str,format: TextDataFormat)

   Adds text data to the data object in the format indicated by the specified 

    System.Windows.Forms.TextDataFormat value.

  

  

   textData: The text to add to the data object.

   format: One of the System.Windows.Forms.TextDataFormat values.

  SetText(self: DataObject,textData: str)

   Adds text data to the data object in the System.Windows.Forms.TextDataFormat.UnicodeText format.

  

   textData: The text to add to the data object.
  """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 @staticmethod
 def __new__(self,*__args):
  """
  __new__(cls: type)

  __new__(cls: type,data: object)

  __new__(cls: type,format: str,data: object)
  """
  pass
 def __repr__(self,*args):
  """ __repr__(self: object) -> str """
  pass



class ICollectData:
 """ Prepares performance data for the performance DLL the system loads when working with performance counters. """
 def CloseData(self):
  """
  CloseData(self: ICollectData)

   Called by the performance DLL's close performance data function.
  """
  pass
 def CollectData(self,id,valueName,data,totalBytes,res):
  """
  CollectData(self: ICollectData,id: int,valueName: IntPtr,data: IntPtr,totalBytes: int) -> IntPtr

  

   Collects the performance data for the performance DLL.

  

   id: The call index.

   valueName: A pointer to a Unicode string list with the requested object identifiers.

   data: A pointer to the data buffer.

   totalBytes: A pointer to a number of bytes.
  """
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass



class SplitterEventArgs(EventArgs):
 """
 Provides data for System.Windows.Forms.Splitter.SplitterMoving and the System.Windows.Forms.Splitter.SplitterMoved events.

 

 SplitterEventArgs(x: int,y: int,splitX: int,splitY: int)
 """
 @staticmethod
 def __new__(self,x,y,splitX,splitY):
  """ __new__(cls: type,x: int,y: int,splitX: int,splitY: int) """
  pass
 SplitX=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the x-coordinate of the upper-left corner of the System.Windows.Forms.Splitter (in client coordinates).



Get: SplitX(self: SplitterEventArgs) -> int



Set: SplitX(self: SplitterEventArgs)=value

"""

 SplitY=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the y-coordinate of the upper-left corner of the System.Windows.Forms.Splitter (in client coordinates).



Get: SplitY(self: SplitterEventArgs) -> int



Set: SplitY(self: SplitterEventArgs)=value

"""

 X=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the x-coordinate of the mouse pointer (in client coordinates).



Get: X(self: SplitterEventArgs) -> int



"""

 Y=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the y-coordinate of the mouse pointer (in client coordinates).



Get: Y(self: SplitterEventArgs) -> int



"""





class SearchForVirtualItemEventArgs(EventArgs):
 """
 Provides data for the System.Windows.Forms.ListView.SearchForVirtualItem event.

 

 SearchForVirtualItemEventArgs(isTextSearch: bool,isPrefixSearch: bool,includeSubItemsInSearch: bool,text: str,startingPoint: Point,direction: SearchDirectionHint,startIndex: int)
 """
 @staticmethod
 def __new__(self,isTextSearch,isPrefixSearch,includeSubItemsInSearch,text,startingPoint,direction,startIndex):
  """ __new__(cls: type,isTextSearch: bool,isPrefixSearch: bool,includeSubItemsInSearch: bool,text: str,startingPoint: Point,direction: SearchDirectionHint,startIndex: int) """
  pass
 Direction=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the direction from the current item that the search should take place.



Get: Direction(self: SearchForVirtualItemEventArgs) -> SearchDirectionHint



"""

 IncludeSubItemsInSearch=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the search should include subitems of list items.



Get: IncludeSubItemsInSearch(self: SearchForVirtualItemEventArgs) -> bool



"""

 Index=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the index of the System.Windows.Forms.ListViewItem found in the System.Windows.Forms.ListView .



Get: Index(self: SearchForVirtualItemEventArgs) -> int



Set: Index(self: SearchForVirtualItemEventArgs)=value

"""

 IsPrefixSearch=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the search should return an item if its text starts with the search text.



Get: IsPrefixSearch(self: SearchForVirtualItemEventArgs) -> bool



"""

 IsTextSearch=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the search is a text search.



Get: IsTextSearch(self: SearchForVirtualItemEventArgs) -> bool



"""

 StartIndex=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the index of the System.Windows.Forms.ListViewItem where the search starts.



Get: StartIndex(self: SearchForVirtualItemEventArgs) -> int



"""

 StartingPoint=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the starting location of the search.



Get: StartingPoint(self: SearchForVirtualItemEventArgs) -> Point



"""

 Text=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the text used to find an item in the System.Windows.Forms.ListView control.



Get: Text(self: SearchForVirtualItemEventArgs) -> str



"""





class CollectionConverter(TypeConverter):
 """
 Provides a type converter to convert collection objects to and from various other representations.

 

 CollectionConverter()
 """
 def ConvertTo(self,*__args):
  """
  ConvertTo(self: CollectionConverter,context: ITypeDescriptorContext,culture: CultureInfo,value: object,destinationType: Type) -> object

  

   Converts the given value object to the specified destination type.

  

   context: An System.ComponentModel.ITypeDescriptorContext that provides a format context.

   culture: The culture to which value will be converted.

   value: The System.Object to convert. This parameter must inherit from System.Collections.ICollection.

   destinationType: The System.Type to convert the value to.

   Returns: An System.Object that represents the converted value.
  """
  pass
 def GetProperties(self,*__args):
  """
  GetProperties(self: CollectionConverter,context: ITypeDescriptorContext,value: object,attributes: Array[Attribute]) -> PropertyDescriptorCollection

  

   Gets a collection of properties for the type of array specified by the value parameter using the 

    specified context and attributes.

  

  

   context: An System.ComponentModel.ITypeDescriptorContext that provides a format context.

   value: An System.Object that specifies the type of array to get the properties for.

   attributes: An array of type System.Attribute that will be used as a filter.

   Returns: A System.ComponentModel.PropertyDescriptorCollection with the properties that are exposed for 

    this data type,or null if there are no properties. This method always returns null.
  """
  pass
 def GetPropertiesSupported(self,context=None):
  """
  GetPropertiesSupported(self: CollectionConverter,context: ITypeDescriptorContext) -> bool

  

   Gets a value indicating whether this object supports properties.

  

   context: An System.ComponentModel.ITypeDescriptorContext that provides a format context.

   Returns: false because 

    System.ComponentModel.CollectionConverter.GetProperties(System.ComponentModel.ITypeDescriptorCont

    ext,System.Object,System.Attribute[]) should not be called to find the properties of this 

    object. This method never returns true.
  """
  pass



class ListViewItemSelectionChangedEventArgs(EventArgs):
 """
 Provides data for the System.Windows.Forms.ListView.ItemSelectionChanged event.

 

 ListViewItemSelectionChangedEventArgs(item: ListViewItem,itemIndex: int,isSelected: bool)
 """
 def __getitem__(self,*args):
  """ x.__getitem__(y) <==> x[y] """
  pass
 @staticmethod
 def __new__(self,item,itemIndex,isSelected):
  """ __new__(cls: type,item: ListViewItem,itemIndex: int,isSelected: bool) """
  pass
 IsSelected=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value indicating whether the item's state has changed to selected.



Get: IsSelected(self: ListViewItemSelectionChangedEventArgs) -> bool



"""

 Item=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the item whose selection state has changed.



Get: Item(self: ListViewItemSelectionChangedEventArgs) -> ListViewItem



"""

 ItemIndex=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the index of the item whose selection state has changed.



Get: ItemIndex(self: ListViewItemSelectionChangedEventArgs) -> int



"""





class TextFormatFlags(Enum,IComparable,IFormattable,IConvertible):
 """
 Specifies the display and layout information for text strings.

 

 enum (flags) TextFormatFlags,values: Bottom (8),Default (0),EndEllipsis (32768),ExpandTabs (64),ExternalLeading (512),GlyphOverhangPadding (0),HidePrefix (1048576),HorizontalCenter (1),Internal (4096),Left (0),LeftAndRightPadding (536870912),ModifyString (65536),NoClipping (256),NoFullWidthCharacterBreak (524288),NoPadding (268435456),NoPrefix (2048),PathEllipsis (16384),PrefixOnly (2097152),PreserveGraphicsClipping (16777216),PreserveGraphicsTranslateTransform (33554432),Right (2),RightToLeft (131072),SingleLine (32),TextBoxControl (8192),Top (0),VerticalCenter (4),WordBreak (16),WordEllipsis (262144)
 """
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==yx.__eq__(y) <==> x==yx.__eq__(y) <==> x==y """
  pass
 def __format__(self,*args):
  """ __format__(formattable: IFormattable,format: str) -> str """
  pass
 def __ge__(self,*args):
  pass
 def __gt__(self,*args):
  pass
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 def __le__(self,*args):
  pass
 def __lt__(self,*args):
  pass
 def __ne__(self,*args):
  pass
 def __reduce_ex__(self,*args):
  pass
 def __str__(self,*args):
  pass
 Bottom=None
 Default=None
 EndEllipsis=None
 ExpandTabs=None
 ExternalLeading=None
 GlyphOverhangPadding=None
 HidePrefix=None
 HorizontalCenter=None
 Internal=None
 Left=None
 LeftAndRightPadding=None
 ModifyString=None
 NoClipping=None
 NoFullWidthCharacterBreak=None
 NoPadding=None
 NoPrefix=None
 PathEllipsis=None
 PrefixOnly=None
 PreserveGraphicsClipping=None
 PreserveGraphicsTranslateTransform=None
 Right=None
 RightToLeft=None
 SingleLine=None
 TextBoxControl=None
 Top=None
 value__=None
 VerticalCenter=None
 WordBreak=None
 WordEllipsis=None




class ComVisibleAttribute(Attribute,_Attribute):
 """
 Controls accessibility of an individual managed type or member,or of all types within an assembly,to COM.

 

 ComVisibleAttribute(visibility: bool)
 """
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 @staticmethod
 def __new__(self,visibility):
  """ __new__(cls: type,visibility: bool) """
  pass
 Value=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets a value that indicates whether the COM type is visible.



Get: Value(self: ComVisibleAttribute) -> bool



"""





# -*- coding: utf-8 -*-
"""Revit-aware event handler."""
#pylint: disable=import-error,invalid-name,broad-except
import sys
import traceback
import threading
import json

from pyrevit.api import UI
from pyrevit.coreutils.logger import get_logger
from pyrevit.coreutils import moduleutils as modutils
from pyrevit.compat import make_request

from pyrevit.routes.server import exceptions as excp
from pyrevit.routes.server import base


mlogger = get_logger(__name__)


ARGS_REQUEST = 'request'
ARGS_UIAPP = 'uiapp'
ARGS_UIDOC = 'uidoc'
ARGS_DOC = 'doc'

RESERVED_VAR_NAMES = [
    ARGS_REQUEST,
    ARGS_UIAPP,
]


class RequestHandler(UI.IExternalEventHandler):
    """Revit external event handler type."""
    # static privates
    _request = None
    _handler = None
    _response = None
    _done = False
    _lock = threading.Lock()

    @property
    def request(self):
        """Get registered request."""
        with self._lock: #pylint: disable=not-context-manager
            return self._request

    @request.setter
    def request(self, request):
        """Set a new request to be passed to handler when event is raised."""
        with self._lock:
            self._request = request

    @property
    def handler(self):
        """Get registered handler."""
        with self._lock: #pylint: disable=not-context-manager
            return self._handler

    @handler.setter
    def handler(self, handler):
        """Set a new handler to be executed when event is raised."""
        if handler and callable(handler):
            with self._lock:
                self._handler = handler
        else:
            raise excp.RouteHandlerIsNotCallableException(handler.__name__)

    @property
    def response(self):
        """Get registered response."""
        with self._lock: #pylint: disable=not-context-manager
            return self._response

    @property
    def done(self):
        """Check if execution of handler is completed and response is set."""
        with self._lock: #pylint: disable=not-context-manager
            return self._done

    def _set_response(self, response):
        with self._lock: #pylint: disable=not-context-manager
            self._response = response
            self._done = True

    def reset(self):
        """Reset internals for new execution."""
        with self._lock: #pylint: disable=not-context-manager
            self._response = None
            self._done = False

    def join(self):
        """Allow other threads to call this method and wait for completion."""
        # wait until handler signals completion
        while True:
            with self._lock:
                if self._done:
                    return

    @staticmethod
    def run_handler(handler, kwargs):
        """Execute the handler function and return base.Response."""
        response = None
        kwargs = kwargs or {}
        if handler and callable(handler):
            try:
                # now call handler, and save response
                response = handler(**kwargs) #pylint: disable=not-callable
            except Exception as hndlr_ex:
                # grab original CLS exception
                clsx = hndlr_ex.clsException #pylint: disable=no-member
                # get exception info
                sys.exc_type, sys.exc_value, sys.exc_traceback = \
                    sys.exc_info()
                # go back one frame to grab exception stack from handler
                # and grab traceback lines
                tb_report = ''.join(
                    traceback.format_tb(sys.exc_traceback)[1:]
                )
                # wrap all the exception info
                response = excp.RouteHandlerException(
                    message=str(hndlr_ex),
                    exception_type=sys.exc_type,
                    exception_traceback=tb_report,
                    clsx_message=clsx.Message,
                    clsx_source=clsx.Source,
                    clsx_stacktrace=clsx.StackTrace,
                    clsx_targetsite=clsx.TargetSite.ToString()
                    )
        else:
            response = \
                excp.RouteHandlerIsNotCallableException(handler.__name__)
        return response

    @staticmethod
    def make_callback(callback_url, response):
        """Prepare request from base.Response and submit to callback url."""
        # parse response object
        r = RequestHandler.parse_response(response)
        # prepare and submit request
        make_request(url=callback_url, headers=r.headers, data=r.data)

    @staticmethod
    def wants_api_context(handler):
        """Check if handler needs host api context."""
        return modutils.has_any_arguments(
            function_obj=handler,
            arg_name_list=[
                ARGS_UIAPP,
                ARGS_UIDOC,
                ARGS_DOC
            ])

    @staticmethod
    def prepare_handler_kwargs(request, handler, uiapp=None):
        """Prepare call arguments for handler function."""
        uidoc = doc = None
        if uiapp:
            uidoc = getattr(uiapp, 'ActiveUIDocument', None)
            if uidoc:
                doc = getattr(uidoc, 'Document', None)

        kwargs = {}
        kwargs[ARGS_REQUEST] = request
        # if route pattern has parameter, provide those as well
        if request.params:
            kwargs.update({x.key:x.value for x in request.params})
        # add host api context params
        kwargs[ARGS_UIAPP] = uiapp
        kwargs[ARGS_UIDOC] = uidoc
        kwargs[ARGS_DOC] = doc

        return modutils.filter_kwargs(handler, kwargs)

    @staticmethod
    def parse_response(response):
        """Parse any given response data and return Response object."""
        status = base.OK
        headers = {}
        data = None

        # can not directly check for isinstance(x, Response)
        # this module is executed on a different Engine than the
        # script that registered the request handler function, thus
        # the Response in script engine does not match Response
        # registered when this module was loaded
        #
        # now process reponse based on obj type
        # it is an exception is has .message
        # write the exeption to output and return
        if hasattr(response, 'message'):
            status = \
                response.status if hasattr(response, 'status') \
                    else base.INTERNAL_SERVER_ERROR
            headers = {'Content-Type': 'application/json'}
            data = json.dumps(
                {
                    "exception": {
                        "source": response.source
                                  if hasattr(response, 'source')
                                  else base.DEFAULT_SOURCE,
                        "message": str(response)
                    }
                }
            )

        # plain text response
        elif isinstance(response, str):
            # keey default status
            headers['Content-Type'] = 'text/html'
            data = json.dumps(response)

        # any obj that has .status and .data, OR
        # any json serializable object
        # serialize before sending results
        # in case exceptions happen in serialization,
        # there are no double status in response header
        else:
            # determine status
            status = getattr(response, 'status', base.OK)

            # determine headers
            headers.update(
                getattr(response, 'headers', {})
                )

            # determine data, or dump the response object
            data = getattr(response, 'data', response)

            # serialize data
            if data is not None:
                data = json.dumps(data)
                headers['Content-Type'] = 'application/json'

        return base.Response(status=status, data=data, headers=headers)

    def Execute(self, uiapp):
        """This method is called to handle the external event."""
        # grab data. getters are thread-safe
        handler = self.handler
        request = self.request
        response = None

        try:
            # process necessary arguments for the handler
            kwargs = RequestHandler.prepare_handler_kwargs(
                request,
                handler,
                uiapp=uiapp
                )
            # run handler with prepared arguments, and grab the response
            response = self.run_handler(handler, kwargs)
        except Exception as exec_ex:
            # create exception response
            response = excp.RouteHandlerExecException(message=str(exec_ex))
        finally:
            # send response to callback url if requested
            if request.callback_url:
                RequestHandler.make_callback(request.callback_url, response)
            # or set the response to be picked up by http request handler
            else:
                self._set_response(response)

    def GetName(self):
        """String identification of the event handler."""
        return self.__class__.__name__



class StructLayoutAttribute(Attribute,_Attribute):
 """
 Lets you control the physical layout of the data fields of a class or structure.

 

 StructLayoutAttribute(layoutKind: LayoutKind)

 StructLayoutAttribute(layoutKind: Int16)
 """
 def __init__(self,*args):
  """ x.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signaturex.__init__(...) initializes x; see x.__class__.__doc__ for signature """
  pass
 @staticmethod
 def __new__(self,layoutKind):
  """
  __new__(cls: type,layoutKind: LayoutKind)

  __new__(cls: type,layoutKind: Int16)
  """
  pass
 Value=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the System.Runtime.InteropServices.LayoutKind value that specifies how the class or structure is arranged.



Get: Value(self: StructLayoutAttribute) -> LayoutKind



"""


 CharSet=None
 Pack=None
 Size=None




class IPAddress(object):
 """
 Provides an Internet Protocol (IP) address.

 

 IPAddress(newAddress: Int64)

 IPAddress(address: Array[Byte],scopeid: Int64)

 IPAddress(address: Array[Byte])
 """
 def Equals(self,comparand):
  """
  Equals(self: IPAddress,comparand: object) -> bool

  

   Compares two IP addresses.

  

   comparand: An System.Net.IPAddress instance to compare to the current instance.

   Returns: true if the two addresses are equal; otherwise,false.
  """
  pass
 def GetAddressBytes(self):
  """
  GetAddressBytes(self: IPAddress) -> Array[Byte]

  

   Provides a copy of the System.Net.IPAddress as an array of bytes.

   Returns: A System.Byte array.
  """
  pass
 def GetHashCode(self):
  """
  GetHashCode(self: IPAddress) -> int

  

   Returns a hash value for an IP address.

   Returns: An integer hash value.
  """
  pass
 @staticmethod
 def HostToNetworkOrder(host):
  """
  HostToNetworkOrder(host: Int16) -> Int16

  

   Converts a short value from host byte order to network byte order.

  

   host: The number to convert,expressed in host byte order.

   Returns: A short value,expressed in network byte order.

  HostToNetworkOrder(host: int) -> int

  

   Converts an integer value from host byte order to network byte order.

  

   host: The number to convert,expressed in host byte order.

   Returns: An integer value,expressed in network byte order.

  HostToNetworkOrder(host: Int64) -> Int64

  

   Converts a long value from host byte order to network byte order.

  

   host: The number to convert,expressed in host byte order.

   Returns: A long value,expressed in network byte order.
  """
  pass
 @staticmethod
 def IsLoopback(address):
  """
  IsLoopback(address: IPAddress) -> bool

  

   Indicates whether the specified IP address is the loopback address.

  

   address: An IP address.

   Returns: true if address is the loopback address; otherwise,false.
  """
  pass
 def MapToIPv4(self):
  """ MapToIPv4(self: IPAddress) -> IPAddress """
  pass
 def MapToIPv6(self):
  """ MapToIPv6(self: IPAddress) -> IPAddress """
  pass
 @staticmethod
 def NetworkToHostOrder(network):
  """
  NetworkToHostOrder(network: Int16) -> Int16

  

   Converts a short value from network byte order to host byte order.

  

   network: The number to convert,expressed in network byte order.

   Returns: A short value,expressed in host byte order.

  NetworkToHostOrder(network: int) -> int

  

   Converts an integer value from network byte order to host byte order.

  

   network: The number to convert,expressed in network byte order.

   Returns: An integer value,expressed in host byte order.

  NetworkToHostOrder(network: Int64) -> Int64

  

   Converts a long value from network byte order to host byte order.

  

   network: The number to convert,expressed in network byte order.

   Returns: A long value,expressed in host byte order.
  """
  pass
 @staticmethod
 def Parse(ipString):
  """
  Parse(ipString: str) -> IPAddress

  

   Converts an IP address string to an System.Net.IPAddress instance.

  

   ipString: A string that contains an IP address in dotted-quad notation for IPv4 and in colon-hexadecimal 

    notation for IPv6.

  

   Returns: An System.Net.IPAddress instance.
  """
  pass
 def ToString(self):
  """
  ToString(self: IPAddress) -> str

  

   Converts an Internet address to its standard notation.

   Returns: A string that contains the IP address in either IPv4 dotted-quad or in IPv6 colon-hexadecimal 

    notation.
  """
  pass
 @staticmethod
 def TryParse(ipString,address):
  """
  TryParse(ipString: str) -> (bool,IPAddress)

  

   Determines whether a string is a valid IP address.

  

   ipString: The string to validate.

   Returns: true if ipString is a valid IP address; otherwise,false.
  """
  pass
 def __eq__(self,*args):
  """ x.__eq__(y) <==> x==y """
  pass
 @staticmethod
 def __new__(self,*__args):
  """
  __new__(cls: type,newAddress: Int64)

  __new__(cls: type,address: Array[Byte],scopeid: Int64)

  __new__(cls: type,address: Array[Byte])
  """
  pass
 def __ne__(self,*args):
  pass
 Address=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """An Internet Protocol (IP) address.



Get: Address(self: IPAddress) -> Int64



Set: Address(self: IPAddress)=value

"""

 AddressFamily=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets the address family of the IP address.



Get: AddressFamily(self: IPAddress) -> AddressFamily



"""

 IsIPv4MappedToIPv6=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Get: IsIPv4MappedToIPv6(self: IPAddress) -> bool



"""

 IsIPv6LinkLocal=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets whether the address is an IPv6 link local address.



Get: IsIPv6LinkLocal(self: IPAddress) -> bool



"""

 IsIPv6Multicast=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets whether the address is an IPv6 multicast global address.



Get: IsIPv6Multicast(self: IPAddress) -> bool



"""

 IsIPv6SiteLocal=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets whether the address is an IPv6 site local address.



Get: IsIPv6SiteLocal(self: IPAddress) -> bool



"""

 IsIPv6Teredo=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets whether the address is an IPv6 Teredo address.



Get: IsIPv6Teredo(self: IPAddress) -> bool



"""

 ScopeId=property(lambda self: object(),lambda self,v: None,lambda self: None)
 """Gets or sets the IPv6 address scope identifier.



Get: ScopeId(self: IPAddress) -> Int64



Set: ScopeId(self: IPAddress)=value

"""


 Any=None
 Broadcast=None
 IPv6Any=None
 IPv6Loopback=None
 IPv6None=None
 Loopback=None
 None=None

